import base64
import contextlib
import logging
import os
import stat
import sys
import tarfile

if sys.version_info >= (3, 14):
    from compression import zstd
else:
    from backports import zstd

from ..archive import Archive, TarfileObjectProcessors, ChunksProcessor, zero_chunk_flags, zero_chunk_id
from ..compress import get_zstd_mt_workers
from ..constants import *  # NOQA
from ..helpers import Error
from ..helpers import HardLinkManager, IncludePatternNeverMatchedWarning
from ..helpers import ProgressIndicatorPercent
from ..helpers import dash_open
from ..helpers import msgpack
from ..helpers import create_filter_process
from ..helpers import ChunkIteratorFileWrapper
from ..helpers import archivename_validator, comment_validator, PathSpec, ChunkerParams, CompressionSpec
from ..helpers import DigestAlgos
from ..helpers import FilesystemPathSpec
from ..helpers import remove_surrogates
from ..helpers import timestamp, archive_ts_now
from ..helpers import basic_json_data, json_print
from ..helpers import log_multi
from ..helpers.argparsing import ArgumentParser
from ..manifest import Manifest

from ._common import with_repository, with_archive, Highlander, define_exclusion_group
from ._common import build_matcher, build_filter

from ..logger import create_logger

logger = create_logger(__name__)

# Python 3.12+ gives a deprecation warning if TarFile.extraction_filter is None.
# https://docs.python.org/3.12/library/tarfile.html#tarfile-extraction-filter
if hasattr(tarfile, "fully_trusted_filter"):
    tarfile.TarFile.extraction_filter = staticmethod(tarfile.fully_trusted_filter)  # type: ignore


def item_to_tarinfo(item, hlm, *, warning=None):
    """Transform a Borg *item* into a tarfile.TarInfo object.

    Return a tuple (tarinfo, needs_content): *tarinfo* is None if the item cannot be
    represented as a TarInfo (and should be skipped), else the TarInfo to write.
    *needs_content* is True when the caller must append the item's file content (its
    chunk data) after the header - this is the case for a regular file that is not a
    tar hard link to an earlier one.

    *hlm* is a HardLinkManager(id_type=bytes, info_type=str) mapping an already-emitted
    file's hlid to its (tar) path, so later hard links become tar LNKTYPE references.
    *warning* is an optional callback(format, *args) for reporting unsupported types.
    """
    tarinfo = tarfile.TarInfo()
    tarinfo.name = item.path
    tarinfo.mtime = item.mtime / 1e9
    tarinfo.mode = stat.S_IMODE(item.mode)
    tarinfo.uid = item.get("uid", 0)
    tarinfo.gid = item.get("gid", 0)
    tarinfo.uname = item.get("user", "")
    tarinfo.gname = item.get("group", "")
    # The linkname in tar has 2 uses:
    # for symlinks it means the destination, while for hard links it refers to the file.
    # Since hard links in tar have a different type code (LNKTYPE) the format might
    # support hardlinking arbitrary objects (including symlinks and directories), but
    # whether implementations actually support that is a whole different question...
    tarinfo.linkname = ""

    needs_content = False
    modebits = stat.S_IFMT(item.mode)
    if modebits == stat.S_IFREG:
        tarinfo.type = tarfile.REGTYPE
        if "hlid" in item:
            linkname = hlm.retrieve(id=item.hlid)
            if linkname is not None:
                # the first hard link was already added to the archive, add a tar-hard-link reference to it.
                tarinfo.type = tarfile.LNKTYPE
                tarinfo.linkname = linkname
            else:
                tarinfo.size = item.get_size()
                needs_content = True
                hlm.remember(id=item.hlid, info=item.path)
        else:
            tarinfo.size = item.get_size()
            needs_content = True
    elif modebits == stat.S_IFDIR:
        tarinfo.type = tarfile.DIRTYPE
    elif modebits == stat.S_IFLNK:
        tarinfo.type = tarfile.SYMTYPE
        tarinfo.linkname = item.target
    elif modebits == stat.S_IFBLK:
        tarinfo.type = tarfile.BLKTYPE
        tarinfo.devmajor = os.major(item.rdev)
        tarinfo.devminor = os.minor(item.rdev)
    elif modebits == stat.S_IFCHR:
        tarinfo.type = tarfile.CHRTYPE
        tarinfo.devmajor = os.major(item.rdev)
        tarinfo.devminor = os.minor(item.rdev)
    elif modebits == stat.S_IFIFO:
        tarinfo.type = tarfile.FIFOTYPE
    else:
        if warning is not None:
            warning("%s: unsupported file type %o for tar export", remove_surrogates(item.path), modebits)
        return None, False
    return tarinfo, needs_content


def item_to_paxheaders(format, item):
    """Transform (parts of) a Borg *item* into a pax_headers dict."""
    # PAX format
    # ----------
    # When using the PAX (POSIX) format, we can support some things that aren't possible
    # with classic tar formats, including GNU tar, such as:
    # - atime, ctime (DONE)
    # - possibly Linux capabilities, security.* xattrs (TODO)
    # - various additions supported by GNU tar in POSIX mode (TODO)
    #
    # BORG format
    # -----------
    # This is based on PAX, but additionally adds BORG.* pax headers.
    # Additionally to the standard tar / PAX metadata and data, it transfers
    # ALL borg item metadata in a BORG specific way.
    #
    ph = {}
    # note: for mtime this is a bit redundant as it is already done by tarfile module,
    #       but we just do it in our way to be consistent for sure.
    for name in "atime", "ctime", "mtime":
        if hasattr(item, name):
            ns = getattr(item, name)
            ph[name] = str(ns / 1e9)
    if hasattr(item, "xattrs"):
        for bkey, bvalue in item.xattrs.items():
            # we have bytes key and bytes value, but the tarfile code
            # expects str key and str value.
            key = SCHILY_XATTR + bkey.decode("utf-8", errors="surrogateescape")
            value = bvalue.decode("utf-8", errors="surrogateescape")
            ph[key] = value
    # Add POSIX access and default ACL if present
    acl_access = item.get("acl_access")
    if acl_access is not None:
        ph[SCHILY_ACL_ACCESS] = acl_access.decode("utf-8", errors="surrogateescape")
    acl_default = item.get("acl_default")
    if acl_default is not None:
        ph[SCHILY_ACL_DEFAULT] = acl_default.decode("utf-8", errors="surrogateescape")
    if format == "BORG":  # BORG format additions
        ph["BORG.item.version"] = "1"
        # BORG.item.meta - just serialize all metadata we have:
        meta_bin = msgpack.packb(item.as_dict())
        meta_text = base64.b64encode(meta_bin).decode()
        ph["BORG.item.meta"] = meta_text
    return ph


def chunks_to_sparse_info(chunks, zero_flags):
    """Compute a GNU sparse map from an item's chunk list.

    *zero_flags* tells for each chunk in *chunks* whether it is an all-zero chunk;
    runs of all-zero chunks become the holes of the sparse file. Holes are shrunk
    to whole 512-byte tar blocks (a trailing hole may end unaligned at the file's
    end): GNU tar's sparse reader processes the data segments block-wise, so it
    desyncs on unaligned segments - and this also matches what GNU tar itself
    produces, as its own hole detection is block-granular. The zero bytes shaved
    off the hole edges become part of the neighboring data segments.

    Return a tuple (map_entries, stream_plan, realsize): *map_entries* is a list of
    (offset, length) data segments - with a terminating (realsize, 0) entry if the
    file ends in a hole, like GNU tar creates it -, *stream_plan* tells how to
    produce the data segments' bytes: a list of ChunkListEntry (a chunk to fetch and
    emit completely) and int (a count of zero bytes to emit literally) elements,
    *realsize* is the logical file size.
    Return None if there is no hole (the file shall be a normal dense tar member).
    """
    if not chunks or all(not zero for zero in zero_flags):
        return None
    realsize = sum(chunk.size for chunk in chunks)
    # runs of all-zero chunks, shrunk to 512-byte block alignment, become the holes.
    holes = []

    def add_hole(start, end):
        start = -(-start // tarfile.BLOCKSIZE) * tarfile.BLOCKSIZE
        if end != realsize:  # a trailing hole may end unaligned, gtar just truncates the file
            end = end // tarfile.BLOCKSIZE * tarfile.BLOCKSIZE
        if end > start:
            holes.append((start, end))

    offset = 0
    run_start = None  # start offset of the current run of all-zero chunks
    for chunk, zero in zip(chunks, zero_flags):
        if zero and run_start is None:
            run_start = offset
        elif not zero and run_start is not None:
            add_hole(run_start, offset)
            run_start = None
        offset += chunk.size
    if run_start is not None:
        add_hole(run_start, realsize)
    if not holes:
        return None  # all zero runs were too short to make a block-aligned hole
    # the data segments are the complement of the holes.
    map_entries = []
    pos = 0
    for hole_start, hole_end in holes:
        if hole_start > pos:
            map_entries.append((pos, hole_start - pos))
        pos = hole_end
    if pos < realsize:
        map_entries.append((pos, realsize - pos))
    else:
        # the file ends in a hole: terminate the map like GNU tar does.
        map_entries.append((realsize, 0))
    # plan the emission of the data segments' bytes, chunk by chunk: data chunks are
    # emitted completely, all-zero chunks only with their parts sticking out of the
    # hole (shaved-off hole edges and the chunks of too-short zero runs) - those are
    # emitted as literal zero bytes, so all-zero chunks never need to be fetched.
    stream_plan = []
    offset = 0
    hole_idx = 0
    for chunk, zero in zip(chunks, zero_flags):
        chunk_end = offset + chunk.size
        if not zero:
            stream_plan.append(chunk)
        else:
            pos = offset
            while pos < chunk_end:
                while hole_idx < len(holes) and holes[hole_idx][1] <= pos:
                    hole_idx += 1
                if hole_idx < len(holes) and holes[hole_idx][0] <= pos:
                    pos = min(holes[hole_idx][1], chunk_end)  # covered by a hole: emit nothing
                else:
                    emit_end = min(holes[hole_idx][0], chunk_end) if hole_idx < len(holes) else chunk_end
                    stream_plan.append(emit_end - pos)
                    pos = emit_end
        offset = chunk_end
    return map_entries, stream_plan, realsize


class SparseTarInfo(tarfile.TarInfo):
    """TarInfo for GNU sparse format 1.0 members.

    A sparse member must never get a pax "size" record: readers recalculate the offset of
    the next header from it *after* the sparse processing already consumed the sparse map,
    so they desync on it - the pax-standard way of representing big sizes is broken for
    sparse members. Thus, the member's stored size (sparse map + data segments) always
    lives in the ustar size field: as standard octal while it fits (< 8 GiB, so those
    members stay as standard-conforming as sparse members can be), else base-256 encoded
    (the GNU/star big-number encoding; not POSIX, but GNU tar does the same for big
    numbers in pax mode, and GNU tar, libarchive/bsdtar and python's tarfile all read it
    in any tar format). The logical file size (GNU.sparse.realsize) is unlimited anyway.

    The base-256 field is patched in manually: python's tarfile writes base-256 only for
    GNU_FORMAT - for PAX_FORMAT it conforms to POSIX and emits a pax "size" record, the
    very thing that must be avoided here.
    """

    __slots__ = ()  # keep the TarInfo object layout, so plain TarInfos can be converted

    octal_size_limit = 8**11  # what fits into the 12-digit octal ustar size field

    def create_pax_header(self, info, encoding):
        stored_size = info["size"]
        if stored_size < self.octal_size_limit:
            return super().create_pax_header(info, encoding)
        info["size"] = 0  # suppresses both the automatic pax "size" record and the octal overflow
        buf = super().create_pax_header(info, encoding)
        # patch the base-256 encoded stored size into the ustar block's size field
        # and recompute the block's checksum.
        ustar = bytearray(buf[-tarfile.BLOCKSIZE :])
        ustar[124:136] = tarfile.itn(stored_size, 12, tarfile.GNU_FORMAT)
        chksum = tarfile.calc_chksums(bytes(ustar))[0]
        ustar[148:156] = bytes("%06o\0 " % chksum, "ascii")
        return buf[: -tarfile.BLOCKSIZE] + bytes(ustar)


def gnu_sparse_10_map(map_entries):
    """Serialize a sparse map as a GNU sparse format 1.0 map block.

    That is a series of decimal numbers delimited by newlines: the number of map
    entries, then offset and length of each entry - NUL-padded to a multiple of
    the 512 byte tar block size. It precedes the data segments in the tar member.
    """
    numbers = [len(map_entries)]
    for offset, length in map_entries:
        numbers += [offset, length]
    text = "".join(f"{number}\n" for number in numbers).encode()
    return text + b"\0" * (-len(text) % tarfile.BLOCKSIZE)


# Sentinel returned by get_tar_filter for zstd suffixes: (de)compress in-process
# via the zstd module instead of piping through an external filter program.
IN_PROCESS_ZSTD = "zstd (in-process)"

# Default zstd compression level, same as the zstd command line tool's default.
ZSTD_TAR_LEVEL = 3


def get_tar_filter(fname, decompress):
    # Note that filter is None if fname is '-'.
    if fname.endswith((".tar.gz", ".tgz")):
        filter = "gzip -d" if decompress else "gzip"
    elif fname.endswith((".tar.bz2", ".tbz")):
        filter = "bzip2 -d" if decompress else "bzip2"
    elif fname.endswith((".tar.xz", ".txz")):
        filter = "xz -d" if decompress else "xz"
    elif fname.endswith((".tar.lz4",)):
        filter = "lz4 -d" if decompress else "lz4"
    elif fname.endswith((".tar.zstd", ".tar.zst", ".tzst")):
        filter = IN_PROCESS_ZSTD
    else:
        filter = None
    logger.debug("Automatically determined tar filter: %s", filter)
    return filter


@contextlib.contextmanager
def create_zstd_filter(stream, stream_close, decompress):
    """In-process zstd (de)compression wrapper around *stream*, same contract as create_filter_process."""
    if decompress:
        zstream = zstd.ZstdFile(stream, "rb")
    else:
        workers = get_zstd_mt_workers(stream=True)
        if workers > 1:
            params = zstd.CompressionParameter
            options = {params.compression_level: ZSTD_TAR_LEVEL, params.nb_workers: workers}
            zstream = zstd.ZstdFile(stream, "wb", options=options)
        else:
            zstream = zstd.ZstdFile(stream, "wb", level=ZSTD_TAR_LEVEL)
    try:
        yield zstream
    except zstd.ZstdError as e:
        raise Error(f"in-process zstd filter failed: {e}") from None
    finally:
        # for compression, closing also finishes the zstd frame.
        # ZstdFile does not close a fileobj it was given, so close stream separately.
        zstream.close()
        if stream_close:
            stream.close()


class TarMixIn:
    @with_repository(compatibility=(Manifest.Operation.READ,))
    @with_archive
    def do_export_tar(self, args, repository, manifest, archive):
        """Export archive contents as a tarball"""
        self.output_list = args.output_list

        if args.sparse and args.tar_format not in ("BORG", "PAX"):
            raise Error("--sparse requires --tar-format BORG or PAX (GNU sparse format 1.0 members are PAX-based).")

        # A quick note about the general design of tar_filter and tarfile;
        # The tarfile module of Python can provide some compression mechanisms
        # by itself, using the built-in gzip, bz2, and lzma modules (and "tar modes"
        # such as "w:xz").
        #
        # Doing so would have three major drawbacks:
        # For one the compressor runs on the same thread as the program using the
        # tarfile, stealing valuable CPU time from Borg and thus reducing throughput.
        # Then this limits the available options - what about lz4? Brotli? zstd?
        # The third issue is that systems can ship more optimized versions than those
        # built into Python, e.g. pigz or pxz, which can use more than one thread for
        # compression.
        #
        # Therefore we externalize compression by using a filter program, which has
        # none of these drawbacks. The only issue of using an external filter is
        # that it has to be installed -- hardly a problem, considering that
        # the decompressor must be installed as well to make use of the exported tarball!
        #
        # zstd is the exception: borg requires the zstd module anyway (Python >= 3.14
        # stdlib, backports.zstd otherwise), libzstd multithreading (see the
        # BORG_ZSTD_MT_WORKERS env var) runs outside the GIL, and there is no
        # significantly more optimized external tool - so zstd tarballs are
        # (de)compressed in-process and need no external program.

        filter = get_tar_filter(args.tarfile, decompress=False) if args.tar_filter == "auto" else args.tar_filter

        tarstream = dash_open(args.tarfile, "wb")
        tarstream_close = args.tarfile != "-"

        if filter is IN_PROCESS_ZSTD:
            filter_context = create_zstd_filter(tarstream, stream_close=tarstream_close, decompress=False)
        else:
            filter_context = create_filter_process(
                filter, stream=tarstream, stream_close=tarstream_close, inbound=False
            )
        with filter_context as _stream:
            self._export_tar(args, archive, _stream)

    def _export_tar(self, args, archive, tarstream):
        # omitting args.pattern_roots here, restricting to paths only by cli args.paths:
        matcher = build_matcher(args.patterns, args.paths)

        progress = args.progress
        output_list = args.output_list
        strip_components = args.strip_components
        hlm = HardLinkManager(id_type=bytes, info_type=str)  # hlid -> path

        filter = build_filter(matcher, strip_components)

        # The | (pipe) symbol instructs tarfile to use a streaming mode of operation
        # where it never seeks on the passed fileobj.
        tar_format = dict(GNU=tarfile.GNU_FORMAT, PAX=tarfile.PAX_FORMAT, BORG=tarfile.PAX_FORMAT)[args.tar_format]
        tar = tarfile.open(fileobj=tarstream, mode="w|", format=tar_format)

        if progress:
            pi = ProgressIndicatorPercent(msg="%5.1f%% Processing: %s", step=0.1, msgid="extract")
            pi.output("Calculating size")
            extracted_size = sum(item.get_size() for item in archive.iter_items(filter))
            pi.total = extracted_size
        else:
            pi = None

        def sparse_chunk_iterator(stream_plan, map_bytes):
            """Generate a sparse member's payload: the sparse map, then the data segments'
            bytes as told by *stream_plan* (chunks to fetch, counts of literal zero bytes)."""
            yield map_bytes
            data_chunks = [entry for entry in stream_plan if not isinstance(entry, int)]
            fetched = archive.pipeline.fetch_many(data_chunks, ro_type=ROBJ_FILE_STREAM)
            for entry in stream_plan:
                yield zeros[:entry] if isinstance(entry, int) else next(fetched)

        def item_content_stream(item, stream_plan=None, map_bytes=None):
            """
            Return a file-like object that reads from the chunks of *item*
            (or produces a sparse member's payload from *stream_plan* / *map_bytes*).
            """
            if stream_plan is not None:
                chunk_iterator = sparse_chunk_iterator(stream_plan, map_bytes)
            else:
                chunk_iterator = archive.pipeline.fetch_many(item.chunks, ro_type=ROBJ_FILE_STREAM)
            if pi:
                info = [remove_surrogates(item.path)]
                return ChunkIteratorFileWrapper(
                    chunk_iterator, lambda read_bytes: pi.show(increase=len(read_bytes), info=info)
                )
            else:
                return ChunkIteratorFileWrapper(chunk_iterator)

        def sparsify_tarinfo(item, tarinfo):
            """Try to turn *tarinfo* into a GNU sparse format 1.0 member.

            If the item's content has detectable holes (runs of all-zero chunks) and storing
            it sparsely is possible and worthwhile, modify *tarinfo* accordingly (mangled
            name, stored size, GNU.sparse.* pax headers) and return (data_chunks, map_bytes);
            else return None and leave *tarinfo* alone (dense member).
            """
            if not item.chunks:
                return None
            ids = [chunk.id for chunk in item.chunks]
            sizes = [chunk.size for chunk in item.chunks]
            # Warm up the zero chunk id memo for likely hole chunk sizes, so such zero chunks
            # get detected even when their id does not repeat within this item: all-zero
            # chunks usually are max-chunk-sized (a power of two, as the chunkers do not cut
            # within runs of zeros) or a file's last chunk (a trailing hole of any size).
            candidate_sizes = {size for size in sizes if size & (size - 1) == 0} | {sizes[-1]}
            for size in candidate_sizes:
                if 0 < size <= len(zeros):
                    zero_chunk_id(archive.key.id_hash, size)
            sparse_info = chunks_to_sparse_info(item.chunks, zero_chunk_flags(ids, sizes, archive.key.id_hash))
            if sparse_info is None:
                return None
            map_entries, stream_plan, realsize = sparse_info
            if realsize != item.get_size():
                return None  # do not write self-contradicting sparse headers for an inconsistent item
            map_bytes = gnu_sparse_10_map(map_entries)
            stored_size = len(map_bytes) + sum(length for _, length in map_entries)
            if stored_size >= realsize:
                return None  # not worthwhile, the map costs more than the holes save
            # SparseTarInfo suppresses the pax "size" record (sparse readers desync on it)
            # and base-256 encodes a stored size beyond the octal ustar field limit.
            tarinfo.__class__ = SparseTarInfo
            # Do like GNU tar: store the member under a mangled name, so that a sparse-unaware
            # tar does not extract the raw map + data segments under the original name - the
            # real name is in the GNU.sparse.name pax header. GNU tar uses its pid where we
            # always use 0, for reproducible output.
            mangled_name = "GNUSparseFile.0/" + tarinfo.name
            # pax record order matters for readers applying them in-order: "path" (mangled)
            # first, the GNU.sparse.* records (real name / size) last, so the latter win.
            ph = {"path": mangled_name}
            ph.update(tarinfo.pax_headers)
            ph["GNU.sparse.major"] = "1"
            ph["GNU.sparse.minor"] = "0"
            ph["GNU.sparse.name"] = tarinfo.name
            ph["GNU.sparse.realsize"] = str(realsize)
            tarinfo.pax_headers = ph
            tarinfo.name = mangled_name
            tarinfo.size = stored_size  # sparse map + data segments, excluding the holes
            return stream_plan, map_bytes

        for item in archive.iter_items(filter):
            orig_path = item.path
            if strip_components:
                item.path = os.sep.join(orig_path.split(os.sep)[strip_components:])
            tarinfo, needs_content = item_to_tarinfo(item, hlm, warning=self.print_warning)
            if tarinfo:
                if args.tar_format in ("BORG", "PAX"):
                    tarinfo.pax_headers = item_to_paxheaders(args.tar_format, item)
                if output_list:
                    logging.getLogger("borg.output.list").info(remove_surrogates(orig_path))
                sparse_content = sparsify_tarinfo(item, tarinfo) if args.sparse and needs_content else None
                if sparse_content is not None:
                    stream_plan, map_bytes = sparse_content
                    stream = item_content_stream(item, stream_plan=stream_plan, map_bytes=map_bytes)
                else:
                    stream = item_content_stream(item) if needs_content else None
                tar.addfile(tarinfo, stream)
                if pi and sparse_content is not None:
                    # the stream callback counted only the stored bytes (sparse map + data
                    # segments), but the progress total is based on the logical file sizes.
                    pi.show(increase=max(0, item.get_size() - tarinfo.size), info=[remove_surrogates(item.path)])

        if pi:
            pi.finish()

        # This does not close the fileobj (tarstream) we passed to it -- a side effect of the | mode.
        tar.close()

        for pattern in matcher.get_unmatched_include_patterns():
            self.print_warning_instance(IncludePatternNeverMatchedWarning(pattern))

    @with_repository(cache=True, compatibility=(Manifest.Operation.WRITE,))
    def do_import_tar(self, args, repository, manifest, cache):
        """Create a backup archive from a tarball"""
        self.output_filter = args.output_filter
        self.output_list = args.output_list

        filter = get_tar_filter(args.tarfile, decompress=True) if args.tar_filter == "auto" else args.tar_filter

        tarstream = dash_open(args.tarfile, "rb")
        tarstream_close = args.tarfile != "-"

        if filter is IN_PROCESS_ZSTD:
            filter_context = create_zstd_filter(tarstream, stream_close=tarstream_close, decompress=True)
        else:
            filter_context = create_filter_process(filter, stream=tarstream, stream_close=tarstream_close, inbound=True)
        with filter_context as _stream:
            self._import_tar(args, repository, manifest, manifest.key, cache, _stream)

    def _import_tar(self, args, repository, manifest, key, cache, tarstream):
        t0 = archive_ts_now()

        archive = Archive(
            manifest,
            args.name,
            cache=cache,
            create=True,
            progress=args.progress,
            chunker_params=args.chunker_params,
            start=t0,
            log_json=args.log_json,
        )
        cp = ChunksProcessor(cache=cache, key=key, add_item=archive.add_item, rechunkify=False)
        tfo = TarfileObjectProcessors(
            cache=cache,
            key=key,
            process_file_chunks=cp.process_file_chunks,
            digest_algos=args.digest_algos,
            add_item=archive.add_item,
            chunker_params=args.chunker_params,
            show_progress=args.progress,
            log_json=args.log_json,
            file_status_printer=self.print_file_status,
        )

        tar = tarfile.open(fileobj=tarstream, mode="r|", ignore_zeros=args.ignore_zeros)

        while tarinfo := tar.next():
            if tarinfo.isreg():
                status = tfo.process_file(tarinfo=tarinfo, status="A", type=stat.S_IFREG, tar=tar)
            elif tarinfo.isdir():
                status = tfo.process_dir(tarinfo=tarinfo, status="d", type=stat.S_IFDIR)
            elif tarinfo.issym():
                status = tfo.process_symlink(tarinfo=tarinfo, status="s", type=stat.S_IFLNK)
            elif tarinfo.islnk():
                # tar uses a hard link model like: the first instance of a hard link is stored as a regular file,
                # later instances are special entries referencing back to the first instance.
                status = tfo.process_hardlink(tarinfo=tarinfo, status="h", type=stat.S_IFREG)
            elif tarinfo.isblk():
                status = tfo.process_dev(tarinfo=tarinfo, status="b", type=stat.S_IFBLK)
            elif tarinfo.ischr():
                status = tfo.process_dev(tarinfo=tarinfo, status="c", type=stat.S_IFCHR)
            elif tarinfo.isfifo():
                status = tfo.process_fifo(tarinfo=tarinfo, status="f", type=stat.S_IFIFO)
            else:
                status = "E"
                self.print_warning("%s: Unsupported tarinfo type %s", tarinfo.name, tarinfo.type)
            self.print_file_status(status, tarinfo.name)

        # This does not close the fileobj (tarstream) we passed to it -- a side effect of the | mode.
        tar.close()

        if args.progress:
            archive.stats.show_progress(final=True)
        archive.stats += tfo.stats
        archive.save(comment=args.comment, timestamp=args.timestamp)
        args.stats |= args.json
        if args.stats:
            if args.json:
                json_print(basic_json_data(archive.manifest, cache=archive.cache, extra={"archive": archive}))
            else:
                log_multi(str(archive), str(archive.stats), logger=logging.getLogger("borg.output.stats"))

    def build_parser_tar(self, subparsers, common_parser, mid_common_parser):
        from ._common import process_epilog

        export_tar_epilog = process_epilog(
            """
        This command creates a tarball from an archive.

        When giving '-' as the output FILE, Borg will write a tar stream to standard output.

        By default (``--tar-filter=auto``) Borg will detect whether the FILE should be compressed
        based on its file extension and pipe the tarball through an appropriate filter
        before writing it to FILE:

        - .tar.gz or .tgz: gzip
        - .tar.bz2 or .tbz: bzip2
        - .tar.xz or .txz: xz
        - .tar.zstd, .tar.zst or .tzst: zstd (in-process, level 3)
        - .tar.lz4: lz4

        For zstd, Borg compresses in-process using libzstd instead of piping through an
        external program (set BORG_ZSTD_MT_WORKERS to use multiple compression threads).
        The other formats are piped through the respective external filter program.

        Alternatively, a ``--tar-filter`` program may be explicitly specified. It should
        read the uncompressed tar stream from stdin and write a compressed/filtered
        tar stream to stdout.

        Depending on the ``--tar-format`` option, these formats are created:

        +--------------+---------------------------+----------------------------+
        | --tar-format | Specification             | Metadata                   |
        +--------------+---------------------------+----------------------------+
        | BORG         | BORG specific, like PAX   | all as supported by borg   |
        +--------------+---------------------------+----------------------------+
        | PAX          | POSIX.1-2001 (pax) format | GNU + atime/ctime/mtime ns |
        |              |                           | + xattrs                   |
        +--------------+---------------------------+----------------------------+
        | GNU          | GNU tar format            | mtime s, no atime/ctime,   |
        |              |                           | no ACLs/xattrs/bsdflags    |
        +--------------+---------------------------+----------------------------+

        With ``--sparse``, files whose content contains runs of all-zero chunks are written
        as sparse tar members (GNU sparse format 1.0, as GNU tar creates it in POSIX mode),
        storing only a hole map and the non-zero data. This requires ``--tar-format BORG``
        or ``PAX``. Such tarballs can be much smaller for sparse files (e.g. disk images)
        and extract to sparse files again with GNU tar's or bsdtar's sparse support
        (as well as with ``borg import-tar`` / ``borg extract --sparse``).
        Notes: hole detection works at the granularity of borg's content chunks (it does not
        depend on the original file having been a sparse file - but some short or unaligned
        zero runs may be stored literally); sparse-unaware tar implementations will extract
        a member as ``GNUSparseFile.0/<name>`` containing the raw hole map and data (the
        same caveat applies to tarballs created by GNU tar); for members needing >= 8 GiB
        of stored (non-hole) data, the stored size is base-256 encoded in the tar header
        (the GNU/star encoding of big numbers, understood by GNU tar, libarchive/bsdtar
        and python) - logical file sizes are unlimited anyway.

        By default the entire archive is extracted but a subset of files and directories
        can be selected by passing a list of ``PATHs`` as arguments.
        The file selection can further be restricted by using the ``--exclude`` option.

        For more help on include/exclude patterns, see the :ref:`borg_patterns` command output.

        ``--progress`` can be slower than no progress display, since it makes one additional
        pass over the archive metadata.
        """
        )
        subparser = ArgumentParser(
            parents=[common_parser], description=self.do_export_tar.__doc__, epilog=export_tar_epilog
        )
        subparsers.add_subcommand("export-tar", subparser, help="create tarball from archive")
        subparser.add_argument(
            "--tar-filter",
            dest="tar_filter",
            default="auto",
            action=Highlander,
            help="filter program to pipe data through",
        )
        subparser.add_argument(
            "--list", dest="output_list", action="store_true", help="output verbose list of items (files, dirs, ...)"
        )
        subparser.add_argument(
            "--tar-format",
            metavar="FMT",
            dest="tar_format",
            default="PAX",
            choices=("BORG", "PAX", "GNU"),
            action=Highlander,
            help="select tar format: BORG, PAX or GNU",
        )
        subparser.add_argument(
            "--sparse",
            dest="sparse",
            action="store_true",
            help="write sparse tar members (GNU sparse format 1.0) for files containing all-zero "
            "chunks (BORG and PAX formats only)",
        )
        subparser.add_argument("name", metavar="NAME", type=archivename_validator, help="specify the archive name")
        subparser.add_argument(
            "tarfile", metavar="FILE", type=FilesystemPathSpec, help='output tar file. "-" to write to stdout instead.'
        )
        subparser.add_argument(
            "paths", metavar="PATH", nargs="*", type=PathSpec, help="paths to extract; patterns are supported"
        )
        define_exclusion_group(subparser, strip_components=True)

        import_tar_epilog = process_epilog(
            """
        This command creates a backup archive from a tarball.

        When giving '-' as path, Borg will read a tar stream from standard input.

        By default (--tar-filter=auto) Borg will detect whether the file is compressed
        based on its file extension and pipe the file through an appropriate filter:

        - .tar.gz or .tgz: gzip -d
        - .tar.bz2 or .tbz: bzip2 -d
        - .tar.xz or .txz: xz -d
        - .tar.zstd, .tar.zst or .tzst: zstd (in-process)
        - .tar.lz4: lz4 -d

        For zstd, Borg decompresses in-process using libzstd instead of piping through
        an external program. The other formats are piped through the respective
        external filter program.

        Alternatively, a --tar-filter program may be explicitly specified. It should
        read compressed data from stdin and output an uncompressed tar stream on
        stdout.

        Most documentation of borg create applies. Note that this command does not
        support excluding files.

        A ``--sparse`` option (as found in borg create) is not needed: sparse members in
        input tarballs (old GNU and PAX sparse formats) are read correctly and their
        holes are stored as deduplicated all-zero chunks.

        About tar formats and metadata conservation or loss, please see ``borg export-tar``.

        import-tar reads these tar formats:

        - BORG: borg specific (PAX-based)
        - PAX: POSIX.1-2001
        - GNU: GNU tar
        - POSIX.1-1988 (ustar)
        - UNIX V7 tar
        - SunOS tar with extended attributes

        To import multiple tarballs into a single archive, they can be simply
        concatenated (e.g. using "cat") into a single file, and imported with an
        ``--ignore-zeros`` option to skip through the stop markers between them.
        """
        )
        subparser = ArgumentParser(
            parents=[common_parser], description=self.do_import_tar.__doc__, epilog=import_tar_epilog
        )
        subparsers.add_subcommand("import-tar", subparser, help=self.do_import_tar.__doc__)
        subparser.add_argument(
            "--tar-filter",
            dest="tar_filter",
            default="auto",
            action=Highlander,
            help="filter program to pipe data through",
        )
        subparser.add_argument(
            "-s",
            "--stats",
            dest="stats",
            action="store_true",
            default=False,
            help="print statistics for the created archive",
        )
        subparser.add_argument(
            "--list",
            dest="output_list",
            action="store_true",
            default=False,
            help="output verbose list of items (files, dirs, ...)",
        )
        subparser.add_argument(
            "--filter",
            dest="output_filter",
            metavar="STATUSCHARS",
            action=Highlander,
            help="only display items with the given status characters",
        )
        subparser.add_argument("--json", action="store_true", help="output stats as JSON (implies --stats)")
        subparser.add_argument(
            "--ignore-zeros",
            dest="ignore_zeros",
            action="store_true",
            help="ignore zero-filled blocks in the input tarball",
        )

        archive_group = subparser.add_argument_group("Archive options")
        archive_group.add_argument(
            "--comment",
            metavar="COMMENT",
            dest="comment",
            type=comment_validator,
            default="",
            action=Highlander,
            help="add a comment text to the archive",
        )
        archive_group.add_argument(
            "--timestamp",
            dest="timestamp",
            type=timestamp,
            default=None,
            action=Highlander,
            metavar="TIMESTAMP",
            help="manually specify the archive creation date/time (yyyy-mm-ddThh:mm:ss[(+|-)HH:MM] format, "
            "(+|-)HH:MM is the UTC offset, default: local time zone). Alternatively, give a reference file/directory.",
        )
        archive_group.add_argument(
            "--chunker-params",
            dest="chunker_params",
            type=ChunkerParams,
            default=CHUNKER_PARAMS,
            action=Highlander,
            metavar="PARAMS",
            help="specify the chunker parameters (ALGO, CHUNK_MIN_EXP, CHUNK_MAX_EXP, "
            "HASH_MASK_BITS, NC_LEVEL). default: %s,%d,%d,%d,%d" % CHUNKER_PARAMS,
        )
        archive_group.add_argument(
            "-C",
            "--compression",
            metavar="COMPRESSION",
            dest="compression",
            type=CompressionSpec,
            default=CompressionSpec("lz4"),
            action=Highlander,
            help="select compression algorithm, see the output of the " '"borg help compression" command for details.',
        )
        archive_group.add_argument(
            "--digests",
            metavar="ALGOS",
            dest="digest_algos",
            action=Highlander,
            type=DigestAlgos,
            default=DIGEST_ALGOS_DEFAULT,
            help="compute these hash digests over the full content of each file and store them into "
            'the archive items. Comma-separated list of hash algorithm names, e.g. "blake3", or "none". '
            "default: %s" % (",".join(DIGEST_ALGOS_DEFAULT) or "none"),
        )

        subparser.add_argument("name", metavar="NAME", type=archivename_validator, help="specify the archive name")
        subparser.add_argument(
            "tarfile",
            metavar="TARFILE",
            type=FilesystemPathSpec,
            help='input tar file. "-" to read from stdin instead.',
        )



# cython: language_level=3

import os
import cython

from cpython.bytes cimport PyBytes_AsString
from libc.stdint cimport uint8_t, uint64_t, int64_t
from libc.string cimport memset

from .phte_chunker cimport ChunkerPHTE

from ..crypto.low_level import CSPRNG


# rabin-aes: a "Chk-PHTE" style content-defined chunker (UHF-then-PRF).
#
# Motivated by "Breaking and Fixing Content-Defined Chunking" (Truong, Merz,
# Scarlata, Guenther, Paterson; eprint 2025/558) and Alexeev/Percival/Zhang
# (arXiv 2504.02095): keyed chunkers built by hiding a secret inside a rolling
# hash (secret buzhash/gear tables) admit key-recovery attacks from known
# plaintext, because every observed cut is an equation about the secret.
# The provably sound construction separates the two jobs:
#
#   1. a rolling *universal hash function* (UHF) compresses the last w bytes
#      into a short digest; here: a Rabin fingerprint over GF(2)[x]/P(x) with
#      a secret random irreducible polynomial P of degree 64, over a 64-byte
#      window (the same UHF family restic uses, with a bigger polynomial).
#      P's degree-64 coefficient is implicit: the digest fills a uint64 and
#      the C kernel's left shifts naturally discard the reduced-away bits.
#   2. a *PRF* (AES-128 with a secret key) is applied to the digest; the cut
#      decision looks only at the PRF output (low bits of the first 8
#      ciphertext bytes, little-endian): cut iff (AES(digest) & mask) == 0.
#
# An attacker who observes cut positions sees only PRF outputs; the digest and
# P remain hidden behind AES. Two windows colliding under the UHF is the only
# residual leakage (they necessarily get equal decisions); for a degree-64
# random irreducible P the collision probability of two distinct 64-byte
# windows is <= 511/2^63 ~= 2^-54.
#
# Security notes (honest limits):
#  * The FSWC-RoR proof bound (eprint 2025/558, Thm 4.5 analogue) degrades
#    with the square of the number of processed positions; with a 64-bit
#    digest it stays meaningful up to roughly tens of GiB per chunker key.
#    Beyond that, no *attack* is known - but the guarantee is heuristic, like
#    everything else in this area except a wider-digest construction.
#  * As with any deterministic chunker, equal chunks still produce equal
#    lengths (deduplication requires this); see the paper's discussion.
#
# Performance: the digest stream is independent of where cuts happen (each
# digest is a pure function of the last 64 bytes), so digests can be computed
# ahead and encrypted in batches. The C kernel (rabin_aes_impl.c) does this
# either via OpenSSL EVP AES-128-ECB batches (portable) or, where available,
# via AES instructions (arm64 crypto extension / x86-64 AES-NI) interleaved
# with the rolling-hash chain. All paths produce bit-identical cuts.

cdef extern from "rabin_aes_impl.h":
    ctypedef struct RA_CTX:
        pass
    int RA_TABLES
    RA_CTX *ra_new(const uint64_t *tables, const uint8_t *aes_key, int kernel)
    int phte_kernel_select(const char *name, int *out_id)
    const char *phte_kernel_names()
    int phte_kernel_default()
    int PHTE_K_EVP
    void ra_free(RA_CTX *ctx)
    const char *ra_kind(const RA_CTX *ctx)
    uint64_t ra_digest64(const RA_CTX *ctx, const uint8_t *q)
    int64_t ra_scan(RA_CTX *ctx, const uint8_t *p, size_t n, uint64_t *digest, uint64_t mask) nogil


# --- GF(2)[x] polynomial helpers (pure Python, used at chunker setup only) ---

_DEG = 64  # degree of the secret irreducible polynomial P (top bit implicit in C)
_WINDOW_SIZE = 64  # bytes; window polynomial degree is 8 * 64 - 1 = 511

# spread table for GF(2) squaring: 8 bits -> 16 bits with zeros interleaved
_SPREAD8 = []
for _b in range(256):
    _s = 0
    for _i in range(8):
        if _b & (1 << _i):
            _s |= 1 << (2 * _i)
    _SPREAD8.append(_s)


def _poly_mod(v, p):
    """v mod p in GF(2)[x]; p must have its top bit set at degree deg(p)."""
    dp = p.bit_length() - 1
    while v.bit_length() - 1 >= dp:
        v ^= p << (v.bit_length() - 1 - dp)
    return v


def _poly_sq_mod(a, p):
    """a^2 mod p in GF(2)[x] (square = interleave zeros between bits)."""
    sq = 0
    shift = 0
    while a:
        sq |= _SPREAD8[a & 0xFF] << shift
        a >>= 8
        shift += 16
    return _poly_mod(sq, p)


def _poly_gcd(a, b):
    """gcd(a, b) in GF(2)[x]."""
    while b:
        a, b = b, _poly_mod(a, b)
    return a


def _is_irreducible(p):
    """Rabin's irreducibility test for deg(p) == 64 (prime factors of 64: 2).

    p is irreducible iff x^(2^64) == x (mod p) and gcd(x^(2^(64/2)) - x, p) == 1,
    i.e. for the exponent 32.
    """
    assert p.bit_length() - 1 == _DEG
    x = 2  # the polynomial "x"
    r = x
    saved = {}
    for i in range(1, _DEG + 1):
        r = _poly_sq_mod(r, p)
        if i == 32:
            saved[i] = r
    if r != x:
        return False
    if _poly_gcd(p, saved[32] ^ x) != 1:
        return False
    return True


def _sample_polynomial(rng):
    """Sample a random irreducible polynomial of degree 64 from the CSPRNG.

    The polynomial has its degree-64 and constant coefficients set; the 63
    middle coefficients are random. About 1 in 64 candidates is irreducible.
    """
    while True:
        v = int.from_bytes(rng.random_bytes(8), "little")
        p = (1 << _DEG) | (v & ((1 << _DEG) - 2)) | 1
        if _is_irreducible(p):
            return p


def _build_tables(p):
    """Build the rolling tables for P (order must match RA_TABLES in the C kernel).

    out_tbl[b]   = poly(b) * x^504 mod P  (remove the byte leaving the window;
                                           504 = 8 * (window_size - 1))
    red_tbl[t]   = poly(t) * x^64 mod P   (reduce the 8 bits shifted above bit 63)
    w1_tbl[t]    = poly(t) * x^72 mod P   (stride-2 step: reduce bits 72..79)
    out8_tbl[b]  = poly(b) * x^512 mod P  (stride-2 removal, newer byte)
    out16_tbl[b] = poly(b) * x^520 mod P  (stride-2 removal, older byte)

    The last three are algebraic combinations of the first two; they let the C
    kernel run two independent even/odd rolling lanes (see rabin_aes_impl.c)
    without changing any cut point.
    """
    out_tbl = [_poly_mod(b << (8 * (_WINDOW_SIZE - 1)), p) for b in range(256)]
    red_tbl = [_poly_mod(t << _DEG, p) for t in range(256)]
    w1_tbl = [_poly_mod(t << (_DEG + 8), p) for t in range(256)]
    out8_tbl = [_poly_mod(b << (8 * _WINDOW_SIZE), p) for b in range(256)]
    out16_tbl = [_poly_mod(b << (8 * (_WINDOW_SIZE + 1)), p) for b in range(256)]
    return out_tbl, red_tbl, w1_tbl, out8_tbl, out16_tbl


def _derive(bytes key):
    """Derive (aes_key, P, tables) deterministically from a 256-bit key.

    Frozen derivation order (changing it changes all cut points and thus breaks
    deduplication against existing repos): first the 16-byte AES key, then P
    (8 bytes per candidate until irreducible).
    """
    rng = CSPRNG(key)
    aes_key = rng.random_bytes(16)
    p = _sample_polynomial(rng)
    tables = _build_tables(p)
    return aes_key, p, tables


from .kernel_env import kernel_error, requested_kernel


cdef int _select_kernel() except -1:
    """Resolve BORG_AES_CHUNKER_KERNEL to a scan path id, raising if it cannot be honoured.

    One selector for all three AES chunkers: they share phte_scan.h, so the
    available paths never differ between them.

    Unset means the fastest path this build and CPU support, see
    phte_kernel_default().
    """
    cdef int kid
    cdef int rc
    want = requested_kernel("BORG_AES_CHUNKER_KERNEL")
    if want is None:
        return phte_kernel_default()
    rc = phte_kernel_select(want.encode("ascii"), &kid)
    if rc != 0:
        raise kernel_error("BORG_AES_CHUNKER_KERNEL", want, rc,
                           (<bytes>phte_kernel_names()).decode("ascii"))
    return kid


cdef class ChunkerRabinAES(ChunkerPHTE):
    """
    Content-Defined Chunker, variable chunk sizes, UHF-then-PRF cut decision.

    Rolling Rabin fingerprint (secret irreducible polynomial, 64-byte window)
    post-processed with AES-128; the cut decision only sees the AES output,
    so observed chunk boundaries do not leak usable information about the
    chunking secret (see module docstring).
    """
    cdef RA_CTX* ctx

    def __cinit__(self, bytes key, int chunk_min_exp, int chunk_max_exp, int hash_mask_bits, int nc_level=0, size_t normal_size=0, bint sparse=False):
        cdef uint64_t c_tables[5 * 256]
        cdef int i, t
        self.ctx = NULL
        self._setup("rabin-aes", chunk_min_exp, chunk_max_exp, hash_mask_bits, nc_level, normal_size, sparse)

        aes_key, p, tables = _derive(key)
        assert len(tables) == RA_TABLES
        for t in range(len(tables)):
            for i in range(256):
                c_tables[t * 256 + i] = tables[t][i]

        kernel = _select_kernel()
        self.ctx = ra_new(c_tables, aes_key, kernel)
        if self.ctx == NULL:
            raise MemoryError("Failed to set up rabin-aes kernel")
        self.kernel_str = (<bytes>ra_kind(self.ctx)).decode("ascii")

    def __dealloc__(self):
        if self.ctx != NULL:
            ra_free(self.ctx)
            self.ctx = NULL

    cdef int64_t _scan(self, const uint8_t *p, size_t n, uint64_t *digest, uint64_t mask) noexcept:
        cdef int64_t r
        with nogil:
            r = ra_scan(self.ctx, p, n, digest, mask)
        return r

    cdef uint64_t _digest64(self, const uint8_t *q) noexcept:
        return ra_digest64(self.ctx, q)


def rabin_aes_get_polynomial(bytes key):
    """Get the secret irreducible polynomial P derived from <key> (for tests)."""
    aes_key, p, tables = _derive(key)
    return p


def rabin_aes_get_tables(bytes key):
    """Get the (out_tbl, red_tbl) rolling tables derived from <key> (for tests)."""
    aes_key, p, tables = _derive(key)
    return tables[0], tables[1]



import hashlib
import io
import json
import os
import random
import shutil
import subprocess
import sys
import tarfile

import pytest
from blake3 import blake3

from ... import xattr
from ...archiver.tar_cmds import chunks_to_sparse_info, gnu_sparse_10_map, SparseTarInfo
from ...constants import *  # NOQA
from ...helpers import Error
from ...item import ChunkListEntry
from .. import changedir
from . import assert_dirs_equal, _extract_hardlinks_setup, cmd, open_archive, requires_hardlinks, RK_ENCRYPTION
from . import create_test_files, create_regular_file
from . import generate_archiver_tests
from ...platform import acl_get, acl_set
from ..platform.platform_test import skipif_not_linux, skipif_acls_not_working

pytest_generate_tests = lambda metafunc: generate_archiver_tests(metafunc, kinds="local,binary")  # NOQA


def have_gnutar():
    if not shutil.which("tar"):
        return False
    popen = subprocess.Popen(["tar", "--version"], stdout=subprocess.PIPE)
    stdout, stderr = popen.communicate()
    return b"GNU tar" in stdout


requires_gnutar = pytest.mark.skipif(not have_gnutar(), reason="GNU tar must be installed for this test.")
requires_gzip = pytest.mark.skipif(not shutil.which("gzip"), reason="gzip must be installed for this test.")
requires_zstd = pytest.mark.skipif(not shutil.which("zstd"), reason="zstd must be installed for this test.")

ZSTD_MAGIC = b"\x28\xb5\x2f\xfd"


@requires_gnutar
def test_export_tar(archivers, request):
    archiver = request.getfixturevalue(archivers)
    create_test_files(archiver.input_path)
    os.unlink("input/flagfile")
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "test", "input")
    cmd(archiver, "export-tar", "test", "simple.tar", "--progress", "--tar-format=GNU")
    with changedir("output"):
        # This probably assumes GNU tar. Note: use the -p switch to extract permissions regardless of umask.
        subprocess.check_call(["tar", "xpf", "../simple.tar", "--warning=no-timestamp"])
    assert_dirs_equal("input", "output/input", ignore_flags=True, ignore_xattrs=True, ignore_ns=True)


@requires_gnutar
@requires_gzip
def test_export_tar_gz(archivers, request):
    archiver = request.getfixturevalue(archivers)
    create_test_files(archiver.input_path)
    os.unlink("input/flagfile")
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "test", "input")
    test_list = cmd(archiver, "export-tar", "test", "simple.tar.gz", "--list", "--tar-format=GNU")
    assert "input/file1\n" in test_list
    assert "input/dir2\n" in test_list
    with changedir("output"):
        subprocess.check_call(["tar", "xpf", "../simple.tar.gz", "--warning=no-timestamp"])
    assert_dirs_equal("input", "output/input", ignore_flags=True, ignore_xattrs=True, ignore_ns=True)


@requires_gnutar
@requires_gzip
def test_export_tar_strip_components(archivers, request):
    archiver = request.getfixturevalue(archivers)
    create_test_files(archiver.input_path)
    os.unlink("input/flagfile")
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "test", "input")
    test_list = cmd(archiver, "export-tar", "test", "simple.tar", "--strip-components=1", "--list", "--tar-format=GNU")
    # --list's paths are those before processing with --strip-components
    assert "input/file1\n" in test_list
    assert "input/dir2\n" in test_list
    with changedir("output"):
        subprocess.check_call(["tar", "xpf", "../simple.tar", "--warning=no-timestamp"])
    assert_dirs_equal("input", "output/", ignore_flags=True, ignore_xattrs=True, ignore_ns=True)


@requires_hardlinks
@requires_gnutar
def test_export_tar_strip_components_links(archivers, request):
    archiver = request.getfixturevalue(archivers)
    _extract_hardlinks_setup(archiver)
    cmd(archiver, "export-tar", "test", "output.tar", "--strip-components=2", "--tar-format=GNU")
    with changedir("output"):
        subprocess.check_call(["tar", "xpf", "../output.tar", "--warning=no-timestamp"])
        assert os.stat("hardlink").st_nlink == 2
        assert os.stat("subdir/hardlink").st_nlink == 2
        assert os.stat("aaaa").st_nlink == 2
        assert os.stat("source2").st_nlink == 2


@requires_hardlinks
@requires_gnutar
def test_extract_hardlinks_tar(archivers, request):
    archiver = request.getfixturevalue(archivers)
    _extract_hardlinks_setup(archiver)
    cmd(archiver, "export-tar", "test", "output.tar", "input/dir1", "--tar-format=GNU")
    with changedir("output"):
        subprocess.check_call(["tar", "xpf", "../output.tar", "--warning=no-timestamp"])
        assert os.stat("input/dir1/hardlink").st_nlink == 2
        assert os.stat("input/dir1/subdir/hardlink").st_nlink == 2
        assert os.stat("input/dir1/aaaa").st_nlink == 2
        assert os.stat("input/dir1/source2").st_nlink == 2


def test_import_tar(archivers, request, tar_format="PAX"):
    archiver = request.getfixturevalue(archivers)
    create_test_files(archiver.input_path, create_hardlinks=False)  # hard links become separate files
    os.unlink("input/flagfile")
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", "src", "input")
    cmd(archiver, "export-tar", "src", "simple.tar", f"--tar-format={tar_format}")
    cmd(archiver, "import-tar", "dst", "simple.tar")
    with changedir(archiver.output_path):
        cmd(archiver, "extract", "dst")
    assert_dirs_equal("input", "output/input", ignore_ns=True, ignore_xattrs=True)


def test_import_tar_nfiles(archivers, request):
    archiver = request.getfixturevalue(archivers)
    # Build a tar with 2 regular files, 1 hardlink, 1 directory and 1 symlink.
    with tarfile.open("input.tar", "w") as tar:
        for name in ("dir/file1", "dir/file2"):
            data = name.encode()
            tarinfo = tarfile.TarInfo(name)
            tarinfo.size = len(data)
            tar.addfile(tarinfo, io.BytesIO(data))
        tarinfo = tarfile.TarInfo("dir/hardlink1")
        tarinfo.type = tarfile.LNKTYPE
        tarinfo.linkname = "dir/file1"
        tar.addfile(tarinfo)
        tarinfo = tarfile.TarInfo("dir/subdir")
        tarinfo.type = tarfile.DIRTYPE
        tar.addfile(tarinfo)
        tarinfo = tarfile.TarInfo("dir/symlink1")
        tarinfo.type = tarfile.SYMTYPE
        tarinfo.linkname = "file1"
        tar.addfile(tarinfo)
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "import-tar", "dst", "input.tar")
    info = json.loads(cmd(archiver, "info", "--json", "dst"))
    # as with borg create, each regular file and each hardlink counts, directories/symlinks do not
    assert info["archives"][0]["stats"]["nfiles"] == 3


def tar_item_digests(archiver, archive):
    """{path: item.digests} as STORED in the items of an archive, see create_cmd_test.item_digests"""
    archive_obj, repository = open_archive(archiver.repository_path, archive)
    with repository:
        return {item.path: item.get("digests") for item in archive_obj.iter_items()}


def test_import_tar_digests(archivers, request):
    # import-tar computes item.digests like borg create does, see #4699
    archiver = request.getfixturevalue(archivers)
    contents = {"dir/file1": os.urandom(512 * 1024), "dir/file2": b"small"}
    with tarfile.open("input.tar", "w") as tar:
        for name, data in contents.items():
            tarinfo = tarfile.TarInfo(name)
            tarinfo.size = len(data)
            tar.addfile(tarinfo, io.BytesIO(data))
        tarinfo = tarfile.TarInfo("dir/hardlink1")  # no content of its own, links to dir/file1
        tarinfo.type = tarfile.LNKTYPE
        tarinfo.linkname = "dir/file1"
        tar.addfile(tarinfo)
        tarinfo = tarfile.TarInfo("dir/subdir")
        tarinfo.type = tarfile.DIRTYPE
        tar.addfile(tarinfo)
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "import-tar", "--digests=blake3", "--chunker-params=fixed,131072", "dst", "input.tar")
    digests = tar_item_digests(archiver, "dst")
    for name, data in contents.items():
        assert digests[name] == {"blake3": blake3(data).digest()}
    # the hard link has no content in the tar, it gets the digests of the file it links to
    assert digests["dir/hardlink1"] == {"blake3": blake3(contents["dir/file1"]).digest()}
    assert digests["dir/subdir"] is None  # a directory has no content and thus no digests

    cmd(archiver, "import-tar", "--digests", "sha256", "dst-sha256", "input.tar")
    digests = tar_item_digests(archiver, "dst-sha256")
    assert digests["dir/file1"] == {"sha256": hashlib.sha256(contents["dir/file1"]).digest()}

    cmd(archiver, "import-tar", "--digests=none", "dst-none", "input.tar")
    assert tar_item_digests(archiver, "dst-none")["dir/file1"] is None

    cmd(archiver, "import-tar", "dst-default", "input.tar")  # digests are opt-in
    assert tar_item_digests(archiver, "dst-default")["dir/file1"] is None


def test_import_unusual_tar(archivers, request):
    archiver = request.getfixturevalue(archivers)

    # Contains these, unusual entries:
    # /foobar
    # ./bar
    # ./foo2/
    # ./foo//bar
    # ./
    tar_archive = os.path.join(os.path.dirname(__file__), "unusual_paths.tar")
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "import-tar", "dst", tar_archive)
    files = cmd(archiver, "list", "dst", "--format", "{path}{NL}").splitlines()
    assert set(files) == {"foobar", "bar", "foo2", "foo/bar", "."}


def test_import_tar_with_dotdot(archivers, request):
    archiver = request.getfixturevalue(archivers)
    if archiver.EXE:  # the test checks for a raised exception. that can't work if the code runs in a separate process.
        pytest.skip("does not work with binaries")

    # Contains this file:
    # ../../../../etc/shadow
    tar_archive = os.path.join(os.path.dirname(__file__), "dotdot_path.tar")
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    with pytest.raises(ValueError, match="unexpected '..' element in path '../../../../etc/shadow'"):
        cmd(archiver, "import-tar", "dst", tar_archive, exit_code=2)


@requires_gzip
def test_import_tar_gz(archivers, request, tar_format="GNU"):
    archiver = request.getfixturevalue(archivers)
    create_test_files(archiver.input_path, create_hardlinks=False)  # hard links become separate files
    os.unlink("input/flagfile")
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", "src", "input")
    cmd(archiver, "export-tar", "src", "simple.tgz", f"--tar-format={tar_format}")
    cmd(archiver, "import-tar", "dst", "simple.tgz")
    with changedir(archiver.output_path):
        cmd(archiver, "extract", "dst")
    assert_dirs_equal("input", "output/input", ignore_ns=True, ignore_xattrs=True)


@pytest.mark.parametrize("suffix", ["tar.zst", "tar.zstd", "tzst"])
def test_export_import_tar_zst(archivers, request, suffix):
    # zstd tarballs are (de)compressed in-process, no external zstd binary is needed.
    archiver = request.getfixturevalue(archivers)
    create_test_files(archiver.input_path, create_hardlinks=False)  # hard links become separate files
    os.unlink("input/flagfile")
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", "src", "input")
    cmd(archiver, "export-tar", "src", f"simple.{suffix}")
    with open(f"simple.{suffix}", "rb") as fd:
        assert fd.read(4) == ZSTD_MAGIC
    cmd(archiver, "import-tar", "dst", f"simple.{suffix}")
    with changedir(archiver.output_path):
        cmd(archiver, "extract", "dst")
    assert_dirs_equal("input", "output/input", ignore_ns=True, ignore_xattrs=True)


def test_export_import_tar_zst_mt(archivers, request, monkeypatch):
    # exercise the multi-threaded compression path (nb_workers) of the in-process zstd filter.
    archiver = request.getfixturevalue(archivers)
    monkeypatch.setenv("BORG_ZSTD_MT_WORKERS", "2")
    monkeypatch.setattr("borg.compress._zstd_mt_workers", None)  # drop the cache
    create_test_files(archiver.input_path, create_hardlinks=False)  # hard links become separate files
    os.unlink("input/flagfile")
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", "src", "input")
    cmd(archiver, "export-tar", "src", "simple.tar.zst")
    cmd(archiver, "import-tar", "dst", "simple.tar.zst")
    with changedir(archiver.output_path):
        cmd(archiver, "extract", "dst")
    assert_dirs_equal("input", "output/input", ignore_ns=True, ignore_xattrs=True)


@requires_gnutar
@requires_zstd
def test_export_tar_zst_interop(archivers, request):
    # in-process zstd output must be readable by the standard zstd tool.
    archiver = request.getfixturevalue(archivers)
    create_test_files(archiver.input_path)
    os.unlink("input/flagfile")
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "test", "input")
    cmd(archiver, "export-tar", "test", "simple.tar.zst", "--tar-format=GNU")
    subprocess.check_call(["zstd", "-d", "simple.tar.zst", "-o", "simple.tar"])
    with changedir("output"):
        subprocess.check_call(["tar", "xpf", "../simple.tar", "--warning=no-timestamp"])
    assert_dirs_equal("input", "output/input", ignore_flags=True, ignore_xattrs=True, ignore_ns=True)


@requires_zstd
def test_tar_filter_zstd_external(archivers, request):
    # an explicit --tar-filter always runs the external filter program, also for zstd.
    archiver = request.getfixturevalue(archivers)
    create_test_files(archiver.input_path, create_hardlinks=False)  # hard links become separate files
    os.unlink("input/flagfile")
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", "src", "input")
    cmd(archiver, "export-tar", "src", "simple.tar.zst", "--tar-filter=zstd")
    with open("simple.tar.zst", "rb") as fd:
        assert fd.read(4) == ZSTD_MAGIC
    cmd(archiver, "import-tar", "dst", "simple.tar.zst", "--tar-filter=zstd -d")
    with changedir(archiver.output_path):
        cmd(archiver, "extract", "dst")
    assert_dirs_equal("input", "output/input", ignore_ns=True, ignore_xattrs=True)


@requires_gnutar
def test_import_concatenated_tar_with_ignore_zeros(archivers, request):
    archiver = request.getfixturevalue(archivers)
    create_test_files(archiver.input_path, create_hardlinks=False)  # hard links become separate files
    os.unlink("input/flagfile")
    with changedir("input"):
        subprocess.check_call(["tar", "cf", "file1.tar", "file1"])
        subprocess.check_call(["tar", "cf", "the_rest.tar", "--exclude", "file1*", "."])
        with open("concatenated.tar", "wb") as concatenated:
            with open("file1.tar", "rb") as file1:
                concatenated.write(file1.read())
            # Clean up for assert_dirs_equal.
            os.unlink("file1.tar")

            with open("the_rest.tar", "rb") as the_rest:
                concatenated.write(the_rest.read())
            # Clean up for assert_dirs_equal.
            os.unlink("the_rest.tar")

    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "import-tar", "--ignore-zeros", "dst", "input/concatenated.tar")
    # Clean up for assert_dirs_equal.
    os.unlink("input/concatenated.tar")

    with changedir(archiver.output_path):
        cmd(archiver, "extract", "dst")
    assert_dirs_equal("input", "output", ignore_ns=True, ignore_xattrs=True)


@requires_gnutar
def test_import_concatenated_tar_without_ignore_zeros(archivers, request):
    archiver = request.getfixturevalue(archivers)
    create_test_files(archiver.input_path, create_hardlinks=False)  # hard links become separate files
    os.unlink("input/flagfile")

    with changedir("input"):
        subprocess.check_call(["tar", "cf", "file1.tar", "file1"])
        subprocess.check_call(["tar", "cf", "the_rest.tar", "--exclude", "file1*", "."])
        with open("concatenated.tar", "wb") as concatenated:
            with open("file1.tar", "rb") as file1:
                concatenated.write(file1.read())
            with open("the_rest.tar", "rb") as the_rest:
                concatenated.write(the_rest.read())
            os.unlink("the_rest.tar")
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "import-tar", "dst", "input/concatenated.tar")

    with changedir(archiver.output_path):
        cmd(archiver, "extract", "dst")
    # Negative test -- assert that only file1 has been extracted, and the_rest has been ignored
    # due to zero-filled block marker.
    assert os.listdir("output") == ["file1"]


@requires_gnutar
def test_import_tar_with_dotslash_paths(archivers, request):
    """Test that paths starting with './' are normalized during import-tar."""
    archiver = request.getfixturevalue(archivers)
    # Create a simple directory structure
    create_regular_file(archiver.input_path, "dir/file")

    # Create a tar file with paths starting with './'
    with changedir("input"):
        # Directly use a path that starts with './'
        subprocess.check_call(["tar", "cf", "dotslash.tar", "./dir"])

        # Verify the tar file contains paths with './' prefix
        tar_content = subprocess.check_output(["tar", "tf", "dotslash.tar"]).decode()
        assert "./dir" in tar_content
        assert "./dir/file" in tar_content

    # Import the tar file into a Borg repository
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "import-tar", "dotslash", "input/dotslash.tar")

    # List the archive contents and verify no paths start with './'
    output = cmd(archiver, "list", "--format={path}{NL}", "dotslash")
    assert "./dir" not in output
    assert "dir" in output
    assert "dir/file" in output


def test_roundtrip_pax_borg(archivers, request):
    archiver = request.getfixturevalue(archivers)
    create_test_files(archiver.input_path)
    os.remove("input/flagfile")  # this would be automagically excluded due to NODUMP
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", "src", "input")
    cmd(archiver, "export-tar", "src", "simple.tar", "--tar-format=BORG")
    cmd(archiver, "import-tar", "dst", "simple.tar")
    with changedir(archiver.output_path):
        cmd(archiver, "extract", "dst")
    assert_dirs_equal("input", "output/input")


def test_roundtrip_pax_xattrs(archivers, request):
    archiver = request.getfixturevalue(archivers)
    if not xattr.is_enabled(archiver.input_path):
        pytest.skip("xattrs not supported")
    create_regular_file(archiver.input_path, "file")
    original_path = os.path.join(archiver.input_path, "file")
    xa_key, xa_value = b"user.xattrtest", b"not valid utf-8: \xff"
    xattr.setxattr(original_path.encode(), xa_key, xa_value)
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", "src", "input")
    cmd(archiver, "export-tar", "src", "xattrs.tar", "--tar-format=PAX")
    cmd(archiver, "import-tar", "dst", "xattrs.tar")
    with changedir(archiver.output_path):
        cmd(archiver, "extract", "dst")
        extracted_path = os.path.abspath("input/file")
        xa_value_extracted = xattr.getxattr(extracted_path.encode(), xa_key)
    assert xa_value_extracted == xa_value


def _sparse_entries(sizes):
    return [ChunkListEntry(id=bytes([i]) * 32, size=size) for i, size in enumerate(sizes)]


def test_chunks_to_sparse_info():
    # no chunks / no holes: not sparse
    assert chunks_to_sparse_info([], []) is None
    assert chunks_to_sparse_info(_sparse_entries([512, 1024]), [False, False]) is None
    # a zero run shorter than a tar block cannot make a (block-aligned) hole
    assert chunks_to_sparse_info(_sparse_entries([512, 511, 512]), [False, True, False]) is None
    # leading hole; adjacent zero chunks coalesce into one hole
    chunks = _sparse_entries([1024, 512, 1536])
    map_entries, stream_plan, realsize = chunks_to_sparse_info(chunks, [True, True, False])
    assert map_entries == [(1536, 1536)]
    assert stream_plan == [chunks[2]]
    assert realsize == 3072
    # middle hole
    chunks = _sparse_entries([512, 1024, 1024, 512])
    map_entries, stream_plan, realsize = chunks_to_sparse_info(chunks, [False, True, True, False])
    assert map_entries == [(0, 512), (2560, 512)]
    assert stream_plan == [chunks[0], chunks[3]]
    assert realsize == 3072
    # trailing hole: terminating (realsize, 0) entry, like GNU tar creates it
    chunks = _sparse_entries([512, 1024])
    map_entries, stream_plan, realsize = chunks_to_sparse_info(chunks, [False, True])
    assert map_entries == [(0, 512), (1536, 0)]
    assert stream_plan == [chunks[0]]
    assert realsize == 1536
    # all-hole file (its trailing hole may end unaligned)
    chunks = _sparse_entries([512, 100])
    map_entries, stream_plan, realsize = chunks_to_sparse_info(chunks, [True, True])
    assert map_entries == [(612, 0)]
    assert stream_plan == []
    assert realsize == 612
    # holes shrink to whole 512-byte tar blocks (GNU tar's sparse reader needs block-aligned
    # data segments); the shaved-off zero bytes at the hole edges are emitted literally.
    chunks = _sparse_entries([10000, 40000, 10000])
    map_entries, stream_plan, realsize = chunks_to_sparse_info(chunks, [False, True, False])
    assert map_entries == [(0, 10240), (49664, 10336)]
    assert stream_plan == [chunks[0], 240, 336, chunks[2]]
    assert realsize == 60000
    # the stream plan produces exactly the data segments' bytes
    emitted = sum(entry if isinstance(entry, int) else entry.size for entry in stream_plan)
    assert emitted == sum(length for _, length in map_entries)


def test_gnu_sparse_10_map():
    # exactly the numbers/layout GNU tar writes, NUL-padded to a multiple of 512.
    map_bytes = gnu_sparse_10_map([(0, 1048576), (5242880, 1048576), (10485760, 0)])
    assert map_bytes == b"3\n0\n1048576\n5242880\n1048576\n10485760\n0\n" + b"\0" * (512 - 39)
    assert len(gnu_sparse_10_map([(1000, 0)])) == 512
    # a large map spills into multiple 512 byte blocks
    map_bytes = gnu_sparse_10_map([(i * 10000, 5000) for i in range(100)])
    assert len(map_bytes) % 512 == 0 and len(map_bytes) > 512
    numbers = [int(n) for n in map_bytes.rstrip(b"\0").split()]
    assert numbers[0] == 100 and numbers[1:3] == [0, 5000]


# chunkers used by the sparse tests: a fixed chunker (all-zero chunks exactly aligned with
# the zero runs) and a small-target fastcdc chunker (content-defined chunk boundaries, so the
# zero runs yield multiple repeated pure all-zero chunks of max. chunk size (64 KiB), possibly
# surrounded by mixed data/zeros chunks at the edges - like real sparse files chunked by the
# default chunker, just scaled down to small test files).
SPARSE_CHUNKER_FIXED = "--chunker-params=fixed,65536"
SPARSE_CHUNKER_CDC = "--chunker-params=fastcdc,12,16,14,2"  # 4 KiB min, 16 KiB target, 64 KiB max


def _create_sparse_test_input(input_path):
    """Create files containing runs of all-zero chunks (the files need not be sparse on disk)."""
    B = 65536  # the max. chunk size of the chunkers used by the sparse tests
    rnd = random.Random(42)  # pseudorandom data, so the cdc chunker cuts realistic chunks
    contents = {
        "sparse_img": rnd.randbytes(B) + b"\0" * (4 * B) + rnd.randbytes(B) + b"\0" * (4 * B),
        "allzero": b"\0" * (3 * B),
        "unaligned": b"\0" * (2 * B) + b"tail",  # leading hole, small unaligned trailing data chunk
        "dense": rnd.randbytes(B + 42),
        "empty": b"",
    }
    for name, data in contents.items():
        create_regular_file(input_path, name, contents=data)
    return contents


@pytest.mark.parametrize("chunker_params", [SPARSE_CHUNKER_FIXED, SPARSE_CHUNKER_CDC])
def test_export_tar_sparse(archivers, request, chunker_params):
    archiver = request.getfixturevalue(archivers)
    contents = _create_sparse_test_input(archiver.input_path)
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", chunker_params, "test", "input")
    cmd(archiver, "export-tar", "test", "dense.tar")
    cmd(archiver, "export-tar", "--sparse", "test", "sparse.tar", "--progress")
    # storing the holes as sparse members must save space (~832 KiB of holes here)
    assert os.path.getsize("sparse.tar") < os.path.getsize("dense.tar") - 500000
    expected_sparse = {"input/sparse_img", "input/allzero", "input/unaligned"}
    seen = set()
    with tarfile.open("sparse.tar") as tar:
        for tarinfo in tar.getmembers():
            if not tarinfo.isreg():
                continue
            seen.add(tarinfo.name)
            # the tarinfo has the real (unmangled) name and the logical size,
            # and the member expands to the original content.
            name = tarinfo.name.rsplit("/", 1)[-1]
            assert tarinfo.size == len(contents[name])
            assert tar.extractfile(tarinfo).read() == contents[name]
            assert tarinfo.issparse() == (tarinfo.name in expected_sparse)
    assert seen == {"input/" + name for name in contents}


@pytest.mark.parametrize("chunker_params", [SPARSE_CHUNKER_FIXED, SPARSE_CHUNKER_CDC])
@pytest.mark.parametrize("tar_format", ["PAX", "BORG"])
def test_export_tar_sparse_roundtrip(archivers, request, tar_format, chunker_params):
    archiver = request.getfixturevalue(archivers)
    _create_sparse_test_input(archiver.input_path)
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", chunker_params, "src", "input")
    cmd(archiver, "export-tar", "--sparse", f"--tar-format={tar_format}", "src", "sparse.tar")
    cmd(archiver, "import-tar", "dst", "sparse.tar")
    with changedir(archiver.output_path):
        cmd(archiver, "extract", "dst")
    assert_dirs_equal("input", "output/input", ignore_ns=True, ignore_xattrs=True)


def test_export_tar_sparse_not_worthwhile(archivers, request):
    # when the sparse map would cost more space than the holes save, store a dense member.
    archiver = request.getfixturevalue(archivers)
    contents = b"X" * 448 + b"\0" * 64
    create_regular_file(archiver.input_path, "tinyhole", contents=contents)
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", "--chunker-params=fixed,64", "test", "input")
    cmd(archiver, "export-tar", "--sparse", "test", "sparse.tar")
    with tarfile.open("sparse.tar") as tar:
        tarinfo = tar.getmember("input/tinyhole")
        assert not tarinfo.issparse()
        assert tar.extractfile(tarinfo).read() == contents


@requires_gnutar
def test_export_tar_sparse_gnutar(archivers, request):
    archiver = request.getfixturevalue(archivers)
    _create_sparse_test_input(archiver.input_path)
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", SPARSE_CHUNKER_CDC, "test", "input")
    cmd(archiver, "export-tar", "--sparse", "test", "sparse.tar")
    with changedir("output"):
        subprocess.check_call(["tar", "xpf", "../sparse.tar", "--warning=no-timestamp"])
    assert_dirs_equal("input", "output/input", ignore_flags=True, ignore_xattrs=True, ignore_ns=True)
    if sys.platform == "linux":
        # GNU tar recreates the holes when extracting sparse members.
        st = os.stat("output/input/sparse_img")
        assert st.st_blocks * 512 < st.st_size


@requires_hardlinks
def test_export_tar_sparse_hardlinks(archivers, request):
    archiver = request.getfixturevalue(archivers)
    contents = b"\0" * 2 * 65536 + b"data"
    create_regular_file(archiver.input_path, "sparse1", contents=contents)
    os.link(os.path.join(archiver.input_path, "sparse1"), os.path.join(archiver.input_path, "sparse2"))
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", SPARSE_CHUNKER_CDC, "src", "input")
    cmd(archiver, "export-tar", "--sparse", "src", "sparse.tar")
    with tarfile.open("sparse.tar") as tar:
        members = [ti for ti in tar.getmembers() if ti.name.startswith("input/sparse")]
        regs = [ti for ti in members if ti.isreg()]
        lnks = [ti for ti in members if ti.islnk()]
        assert len(regs) == 1 and len(lnks) == 1
        # the first occurrence carries the sparse content, the second one is a tar hard link
        # referencing the first one's real (unmangled) name.
        assert regs[0].issparse()
        assert tar.extractfile(regs[0]).read() == contents
        assert lnks[0].linkname == regs[0].name
    # roundtrip: as usual for import-tar, tar hard links become separate files (sharing chunks).
    cmd(archiver, "import-tar", "dst", "sparse.tar")
    with changedir(archiver.output_path):
        cmd(archiver, "extract", "dst")
        for name in "input/sparse1", "input/sparse2":
            with open(name, "rb") as f:
                assert f.read() == contents


def test_export_tar_sparse_strip_components(archivers, request):
    # the mangled member name and GNU.sparse.name must be based on the stripped path.
    archiver = request.getfixturevalue(archivers)
    contents = b"\0" * 2 * 65536 + b"end"
    create_regular_file(archiver.input_path, "dir/sparsefile", contents=contents)
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", SPARSE_CHUNKER_CDC, "test", "input")
    cmd(archiver, "export-tar", "--sparse", "--strip-components=1", "test", "sparse.tar")
    with tarfile.open("sparse.tar") as tar:
        tarinfo = tar.getmember("dir/sparsefile")
        assert tarinfo.issparse()
        assert tar.extractfile(tarinfo).read() == contents


def test_sparse_tarinfo_base256_size():
    # a stored size beyond the 12-digit octal ustar field limit is base-256 encoded in the
    # ustar size field (with a correct checksum), and no pax "size" record is emitted
    # (sparse readers desync on such a record).
    tarinfo = SparseTarInfo(name="GNUSparseFile.0/big")
    tarinfo.size = 3 * 8**11  # 24 GiB of stored data, does not fit the octal field
    tarinfo.pax_headers = {"GNU.sparse.realsize": str(100 * 8**11)}
    buf = tarinfo.tobuf(tarfile.PAX_FORMAT, tarfile.ENCODING, "surrogateescape")
    assert b" size=" not in buf  # no pax "size" record (" realsize=" does not match)
    # frombuf validates the checksum and decodes the base-256 size field:
    parsed = tarfile.TarInfo.frombuf(buf[-tarfile.BLOCKSIZE :], tarfile.ENCODING, "surrogateescape")
    assert parsed.size == 3 * 8**11
    # small stored sizes keep the plain octal encoding:
    tarinfo.size = 4711
    buf = tarinfo.tobuf(tarfile.PAX_FORMAT, tarfile.ENCODING, "surrogateescape")
    assert buf[-tarfile.BLOCKSIZE :][124:136] == b"00000011147\x00"


def test_export_tar_sparse_base256_size(archivers, request, monkeypatch):
    # end-to-end wire-format test of the base-256 stored size: the encoding does not depend
    # on the value, so force it for small members instead of storing 8 GiB of data.
    archiver = request.getfixturevalue(archivers)
    if archiver.EXE:
        pytest.skip("monkeypatching does not reach a borg binary")
    monkeypatch.setattr(SparseTarInfo, "octal_size_limit", 1)
    contents = _create_sparse_test_input(archiver.input_path)
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", SPARSE_CHUNKER_CDC, "src", "input")
    cmd(archiver, "export-tar", "--sparse", "src", "sparse.tar")
    with tarfile.open("sparse.tar") as tar:
        for tarinfo in tar.getmembers():
            if tarinfo.isreg():
                name = tarinfo.name.rsplit("/", 1)[-1]
                assert tar.extractfile(tarinfo).read() == contents[name]
    if have_gnutar():
        with changedir("output"):
            subprocess.check_call(["tar", "xpf", "../sparse.tar", "--warning=no-timestamp"])
        assert_dirs_equal("input", "output/input", ignore_flags=True, ignore_xattrs=True, ignore_ns=True)
        shutil.rmtree("output/input")
    cmd(archiver, "import-tar", "dst", "sparse.tar")
    with changedir(archiver.output_path):
        cmd(archiver, "extract", "dst")
    assert_dirs_equal("input", "output/input", ignore_ns=True, ignore_xattrs=True)


def test_export_tar_sparse_gnu_format_error(archivers, request):
    # --sparse requires a PAX-based tar format, the GNU format cannot store the sparse headers.
    archiver = request.getfixturevalue(archivers)
    create_regular_file(archiver.input_path, "file", contents=b"x")
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", "test", "input")
    if archiver.FORK_DEFAULT:
        output = cmd(archiver, "export-tar", "--sparse", "--tar-format=GNU", "test", "out.tar", exit_code=2)
        assert "--sparse requires --tar-format" in output
    else:
        with pytest.raises(Error, match="--sparse requires --tar-format"):
            cmd(archiver, "export-tar", "--sparse", "--tar-format=GNU", "test", "out.tar")


@skipif_not_linux
@skipif_acls_not_working
def test_acl_roundtrip(archivers, request):
    """Test the complete workflow for POSIX ACLs with export-tar and import-tar.

    This test follows the workflow:
    1. set filesystem ACLs
    2. create a Borg archive
    3. export-tar this archive
    4. import-tar the resulting tar file
    5. extract the imported archive
    6. check the expected ACLs in the filesystem
    """
    archiver = request.getfixturevalue(archivers)

    # Define helper functions for working with ACLs
    def get_acl(path):
        item = {}
        acl_get(path, item, os.stat(path))
        return item

    def set_acl(path, access=None, default=None):
        item = {"acl_access": access, "acl_default": default}
        acl_set(path, item)

    # Define example ACLs
    ACCESS_ACL = b"user::rw-\nuser:root:rw-:0\ngroup::r--\ngroup:root:r--:0\nmask::rw-\nother::r--"
    DEFAULT_ACL = b"user::rw-\nuser:root:r--:0\ngroup::r--\ngroup:root:r--:0\nmask::rw-\nother::r--"

    # 1. Set filesystem ACLs
    # Create test files with ACLs
    create_regular_file(archiver.input_path, "file")
    os.mkdir(os.path.join(archiver.input_path, "dir"))

    file_path = os.path.join(archiver.input_path, "file")
    dir_path = os.path.join(archiver.input_path, "dir")

    # Set ACLs on the test files
    try:
        set_acl(file_path, access=ACCESS_ACL)
        set_acl(dir_path, access=ACCESS_ACL, default=DEFAULT_ACL)
    except OSError as e:
        pytest.skip(f"Failed to set ACLs: {e}")

    file_acl = get_acl(file_path)
    dir_acl = get_acl(dir_path)

    if not file_acl.get("acl_access") or not dir_acl.get("acl_access") or not dir_acl.get("acl_default"):
        pytest.skip("ACLs not supported or not working correctly")

    # 2. Create a Borg archive
    cmd(archiver, "repo-create", "--encryption=none-sha256")
    cmd(archiver, "create", "original", "input")

    # 3. export-tar this archive to a tar file
    cmd(archiver, "export-tar", "original", "acls.tar", "--tar-format=PAX")

    # 4. import-tar the resulting tar file
    cmd(archiver, "import-tar", "imported", "acls.tar")

    # 5. Extract the imported archive
    with changedir(archiver.output_path):
        cmd(archiver, "extract", "imported")

        # 6. Check the expected ACLs in the filesystem
        extracted_file_path = os.path.abspath("input/file")
        extracted_dir_path = os.path.abspath("input/dir")

        extracted_file_acl = get_acl(extracted_file_path)
        extracted_dir_acl = get_acl(extracted_dir_path)

        # Check that access ACLs were preserved
        assert "acl_access" in extracted_file_acl
        assert extracted_file_acl["acl_access"] == file_acl["acl_access"]
        assert b"user:root:rw-" in file_acl["acl_access"]

        assert "acl_access" in extracted_dir_acl
        assert extracted_dir_acl["acl_access"] == dir_acl["acl_access"]
        assert b"user:root:rw-" in dir_acl["acl_access"]

        # Check that default ACLs were preserved for directories
        assert "acl_default" in extracted_dir_acl
        assert extracted_dir_acl["acl_default"] == dir_acl["acl_default"]
        assert b"user:root:r--" in dir_acl["acl_default"]



from .xattr import _listxattr_inner, _getxattr_inner, _setxattr_inner, split_lstring



cdef extern from "sys/extattr.h":
    ssize_t c_extattr_list_file "extattr_list_file" (const char *path, int attrnamespace, void *data, size_t nbytes)
    ssize_t c_extattr_list_link "extattr_list_link" (const char *path, int attrnamespace, void *data, size_t nbytes)
    ssize_t c_extattr_list_fd "extattr_list_fd" (int fd, int attrnamespace, void *data, size_t nbytes)

    ssize_t c_extattr_get_file "extattr_get_file" (const char *path, int attrnamespace, const char *attrname, void *data, size_t nbytes)
    ssize_t c_extattr_get_link "extattr_get_link" (const char *path, int attrnamespace, const char *attrname, void *data, size_t nbytes)
    ssize_t c_extattr_get_fd "extattr_get_fd" (int fd, int attrnamespace, const char *attrname, void *data, size_t nbytes)

    int c_extattr_set_file "extattr_set_file" (const char *path, int attrnamespace, const char *attrname, const void *data, size_t nbytes)
    int c_extattr_set_link "extattr_set_link" (const char *path, int attrnamespace, const char *attrname, const void *data, size_t nbytes)
    int c_extattr_set_fd "extattr_set_fd" (int fd, int attrnamespace, const char *attrname, const void *data, size_t nbytes)

    int EXTATTR_NAMESPACE_USER


# On NetBSD, Borg currently only deals with the USER namespace, as it is unclear
# whether (and, if so, how exactly) it should deal with the SYSTEM namespace.
NS_ID_MAP = {b"user": EXTATTR_NAMESPACE_USER, }


def split_ns(ns_name, default_ns):
    # Split ns_name (which is in the form b"namespace.name") into namespace and name.
    # If there is no namespace given in ns_name, default to default_ns.
    # We also need to deal with "unexpected" namespaces here — they could come
    # from Borg archives made on other operating systems.
    ns_name_tuple = ns_name.split(b".", 1)
    if len(ns_name_tuple) == 2:
        # We have a namespace prefix in the given name.
        ns, name = ns_name_tuple
    else:
        # No namespace given in ns_name (no dot found); maybe data from an old Borg archive.
        ns, name = default_ns, ns_name
    return ns, name


def listxattr(path, *, follow_symlinks=False):
    def func(path, buf, size):
        if isinstance(path, int):
            return c_extattr_list_fd(path, ns_id, <char *> buf, size)
        else:
            if follow_symlinks:
                return c_extattr_list_file(path, ns_id, <char *> buf, size)
            else:
                return c_extattr_list_link(path, ns_id, <char *> buf, size)

    ns = b"user"
    ns_id = NS_ID_MAP[ns]
    n, buf = _listxattr_inner(func, path)
    return [ns + b"." + name for name in split_lstring(buf[:n]) if name]


def getxattr(path, name, *, follow_symlinks=False):
    def func(path, name, buf, size):
        if isinstance(path, int):
            return c_extattr_get_fd(path, ns_id, name, <char *> buf, size)
        else:
            if follow_symlinks:
                return c_extattr_get_file(path, ns_id, name, <char *> buf, size)
            else:
                return c_extattr_get_link(path, ns_id, name, <char *> buf, size)

    ns, name = split_ns(name, b"user")
    ns_id = NS_ID_MAP[ns]  # this will raise a KeyError it the namespace is unsupported
    n, buf = _getxattr_inner(func, path, name)
    return bytes(buf[:n])


def setxattr(path, name, value, *, follow_symlinks=False):
    def func(path, name, value, size):
        if isinstance(path, int):
            return c_extattr_set_fd(path, ns_id, name, <char *> value, size)
        else:
            if follow_symlinks:
                return c_extattr_set_file(path, ns_id, name, <char *> value, size)
            else:
                return c_extattr_set_link(path, ns_id, name, <char *> value, size)

    ns, name = split_ns(name, b"user")
    try:
        ns_id = NS_ID_MAP[ns]  # this will raise a KeyError it the namespace is unsupported
    except KeyError:
        pass
    else:
        _setxattr_inner(func, path, name, value)



import io

from ...platform import swidth, SyncFile


def test_swidth_ascii():
    assert swidth("borg") == 4


def test_swidth_cjk():
    assert swidth("バックアップ") == 6 * 2


def test_swidth_mixed():
    assert swidth("borgバックアップ") == 4 + 6 * 2


def test_syncfile_seek_tell(tmp_path):
    """SyncFile exposes seek() and tell() from the underlying file object."""
    path = tmp_path / "testfile"
    with SyncFile(path, binary=True) as sf:
        sf.write(b"hello world")
        assert sf.tell() == 11
        sf.seek(0, io.SEEK_SET)
        assert sf.tell() == 0
        sf.seek(0, io.SEEK_END)
        assert sf.tell() == 11
        sf.seek(5, io.SEEK_SET)
        assert sf.tell() == 5
        assert sf.read() == b" world"
    assert path.read_bytes() == b"hello world"


def test_syncfile_close_idempotent(tmp_path):
    """Calling SyncFile.close() twice does not raise."""
    path = tmp_path / "testfile"
    sf = SyncFile(path, binary=True)
    sf.write(b"data")
    sf.close()
    sf.close()  # must not raise



import pytest

from ...helpers.nanorst import rst_to_text


def test_inline():
    assert rst_to_text("*foo* and ``bar``.") == "foo and bar."


def test_inline_spread():
    assert rst_to_text("*foo and bar, thusly\nfoobar*.") == "foo and bar, thusly\nfoobar."


def test_comment_inline():
    assert rst_to_text("Foo and Bar\n.. foo\nbar") == "Foo and Bar\n.. foo\nbar"


def test_inline_escape():
    assert rst_to_text('Such as "\\*" characters.') == 'Such as "*" characters.'


def test_comment():
    assert rst_to_text("Foo and Bar\n\n.. foo\nbar") == "Foo and Bar\n\nbar"


def test_directive_note():
    assert rst_to_text(".. note::\n   Note this and that") == "Note:\n   Note this and that"


def test_ref():
    references = {"foo": "baz"}
    assert rst_to_text("See :ref:`fo\no`.", references=references) == "See baz."


def test_undefined_ref():
    with pytest.raises(ValueError) as exc_info:
        rst_to_text("See :ref:`foo`.")
    assert "Undefined reference" in str(exc_info.value)


def test_code_block_end_of_string():
    # check that there is no unexpected exception if a code block
    # is not followed by blank lines, but the string just ends.
    assert rst_to_text("This is a code block::\n\n    borg --help") == "This is a code block:\n\n    borg --help"


def test_code_block_indented_context():
    # a code block inside a definition list: the following text is indented, but less than
    # the code block's contents, so it ends the code block and gets markup-processed again.
    rst = "term\n    Foo::\n\n        code\n\n    Continued ``text``.\n"
    assert rst_to_text(rst) == "term\n    Foo:\n\n        code\n\n    Continued text.\n"


def test_code_block_less_indented_than_four():
    # code blocks indented by less than 4 spaces are recognized, too
    rst = "Foo::\n\n  code\n\nBar ``baz``.\n"
    assert rst_to_text(rst) == "Foo:\n\n  code\n\nBar baz.\n"



from hashlib import sha256
from io import BytesIO


from . import cf
from ...chunkers import ChunkerRabinAES
from ...chunkers.rabin_aes import rabin_aes_get_polynomial, rabin_aes_get_tables, _is_irreducible
from ...constants import *  # NOQA
from ...helpers import hex_to_bin


# from os.urandom(32)
key0 = hex_to_bin("ad9f89095817f0566337dc9ee292fcd59b70f054a8200151f1df5f21704824da")
key1 = hex_to_bin("f1088c7e9e6ae83557ad1558ff36c44a369ea719d1081c29684f52ffccb72cb8")


def H(data):
    return sha256(data).digest()


def test_chunkpoints_rabin_aes_unchanged():
    def twist(size):
        x = 1
        a = bytearray(size)
        for i in range(size):
            x = (x * 1103515245 + 12345) & 0x7FFFFFFF
            a[i] = x & 0xFF
        return a

    data = twist(100000)

    runs = []
    for nc_level in (0, 2, 3):
        for minexp in (6, 7, 11, 12):  # rabin-aes requires minexp >= 6 (64-byte window)
            for maxexp in (15, 17):
                if minexp >= maxexp:
                    continue
                for maskbits in (7, 10, 12):
                    if maskbits - nc_level < 1:  # nc_level needs room below the base mask bits
                        continue
                    for key in (key0, key1):
                        fh = BytesIO(data)
                        chunker = ChunkerRabinAES(key, minexp, maxexp, maskbits, nc_level)
                        chunks = [H(c) for c in cf(chunker.chunkify(fh, -1))]
                        runs.append(H(b"".join(chunks)))

    # The "correct" hash below matches the existing chunker behavior.
    # Future chunker optimizations must not change this, or existing repos will bloat.
    overall_hash = H(b"".join(runs))
    print(overall_hash.hex())
    assert overall_hash == hex_to_bin("f502d8da9dc338ff6949d89cdb545ff690bba8919263a17010c2a4995031d7ef")


def test_rabin_aes_polynomial():
    # P must be a degree-64 irreducible polynomial, deterministically derived from the key
    p0 = rabin_aes_get_polynomial(key0)
    assert p0.bit_length() - 1 == 64
    assert p0 & 1  # constant term set (otherwise divisible by x)
    assert _is_irreducible(p0)
    assert p0 == rabin_aes_get_polynomial(key0)
    p1 = rabin_aes_get_polynomial(key1)
    assert p0 != p1

    out_tbl, red_tbl = rabin_aes_get_tables(key0)
    assert len(out_tbl) == 256 and len(red_tbl) == 256
    for v in out_tbl + red_tbl:
        assert isinstance(v, int)
        assert 0 <= v < 2**64  # all table entries are mod-P remainders
    # spot check the tables against the polynomial: entry b is (b << shift) mod P
    for b in (0, 1, 2, 128, 255):
        v = b << 504
        while v.bit_length() - 1 >= 64:
            v ^= p0 << (v.bit_length() - 1 - 64)
        assert out_tbl[b] == v



#!/usr/bin/env python3
"""Regenerate FAME.md and FAME.svg, a contributor summary produced by git-fame.

FAME.md is the full table, FAME.svg a bar chart of the busiest contributors that
FAME.md embeds.  The chart is drawn here rather than taken from git-fame's own
`--format=svg`, which renders the whole table as monospaced text - fine for a
handful of contributors, unreadable for a few hundred.  It leaves out bots,
which do appear in the table.

Usage: python scripts/fame.py [--top N] [--svg-top N]

Requires `git-fame` (pip install git-fame) and a full (non-shallow) clone,
because the line counts come from `git blame` over the whole history.

The statistics always describe the master branch, no matter which branch the
script is run from, so that FAME.md does not change depending on the checkout.

Contributors are sorted by commit count.  git-fame also reports "surviving"
lines of code, i.e. lines of the current checkout that `git blame` attributes
to an author; that is useful context, but it silently discounts work that was
later refactored away, so it is not what we rank by.

Author identities are merged via .mailmap, so fix duplicates there rather than
here.

Both files are left alone if nothing but the generation date would change, so
that running this regularly (see .github/workflows/fame.yml) does not produce a
diff for a week in which the statistics did not move.
"""

import argparse
import json
import re
import subprocess
from datetime import date
from pathlib import Path
from xml.sax.saxutils import escape

BRANCH = "master"

DATE_LINE = re.compile(r"^Generated on \d{4}-\d{2}-\d{2} ", re.MULTILINE)

# Bots keep their row in the table - they did make those commits - but they
# would take a slot away from a person in the chart, which only has room for a
# few.  GitHub appends "[bot]" to the account name of a bot, e.g.
# "dependabot[bot]"; matching that is safer than guessing from the name.
BOT = re.compile(r"\[bot\]$")

# Generated and binary-ish files would drown out hand-written code.  Excalidraw
# scenes are the worst offenders: they are pretty-printed JSON, so a single
# diagram of a few boxes and arrows is over 10000 lines.  Image formats are
# listed for good measure - git-fame skips binary files by itself, but .svg is
# text and would count.
EXCLUDE = r".*\.(excalidraw|png|jpg|jpeg|gif|svg|ico)$"

HEADERS = ("Contributor", "Commits", "Lines", "Files")

PREAMBLE = """\
# Contributors

{count} people have contributed to Borg, with {commits} commits in total.
Thanks to everyone who helped!

![Contributors by commit count](FAME.svg)

Generated on {date} by `scripts/fame.py`, which computes the statistics with
[git-fame](https://github.com/casperdcl/git-fame), from the `{branch}` branch
only - commits that exist solely on other branches or in unmerged pull requests
are not counted.

This list is only a rough summary: it counts commits and lines, which say
nothing about reviews, bug reports, translations, documentation review, support
work or funding.  See `AUTHORS` for a manually maintained list of contributors.

"Lines" are surviving lines of code: lines of the current master that
`git blame` attributes to that author.  Code that was later rewritten does not
show up there, so the number understates early contributions.  Generated files
(such as the Excalidraw sources of the documentation figures) are excluded.

"""


# Chart geometry, in user units.  The viewport is given in user units plus a
# viewBox rather than in em, because em on the root <svg> resolves against the
# font size that element inherits, not against the font size of the text inside.
SVG_TOP = 20  # a chart of all 300+ contributors would be unreadable
SVG_WIDTH = 760
SVG_ROW_H = 22
SVG_PAD = 14
SVG_LABEL_W = 240  # x where the bars start, i.e. width of the name column
SVG_COUNT_W = 70  # right hand column holding the commit count
SVG_HEAD_H = 58
SVG_FOOT_H = 26
SVG_NAME_MAX = 32  # characters that fit into the name column

# Fills are set per class so that the chart can follow the reader's colour
# scheme; GitHub renders the file in an <img>, where the media query applies.
# The bars use the green of the logo (docs/_static/logo.svg).  That green is
# made for a dark background - on white it only reaches a contrast of 1.9:1,
# below the 3:1 that WCAG asks for graphics - so the light scheme darkens it
# while keeping the hue.
SVG_STYLE = """\
<style>
    text { font-family: -apple-system, "Segoe UI", Helvetica, Arial, sans-serif }
    .bg { fill: #ffffff }
    .title { font-size: 16px; font-weight: 600; fill: #1f2328 }
    .name { font-size: 12px; fill: #1f2328 }
    .subtitle, .footer, .count { font-size: 11px; fill: #59636e }
    .count { font-size: 12px; text-anchor: end }
    .bar { fill: #00a000 }
    @media (prefers-color-scheme: dark) {
        .bg { fill: #0d1117 }
        .title, .name { fill: #e6edf3 }
        .subtitle, .footer, .count { fill: #9198a1 }
        .bar { fill: #00dd00 }
    }
</style>"""


def collect():
    """Run git-fame and return (rows, total), all contributors, most commits first."""
    out = subprocess.check_output(
        ["git-fame", "--silent-progress", "--format=json", "--sort=commits", f"--excl={EXCLUDE}", f"--branch={BRANCH}"],
        text=True,
    )
    fame = json.loads(out)
    rows = [(name, commits, loc, files) for name, loc, commits, files, *_percentages in fame["data"]]
    # git-fame leaves people with the same commit count in its own order, which
    # is stable for a given checkout but not across checkouts, so equal ranks
    # would shuffle from one run to the next.  Break the tie ourselves.
    rows.sort(key=lambda row: (-row[1], -row[2], row[0]))
    return rows, fame["total"]


def render(rows, total, contributors):
    """Return the full FAME.md contents, listing `rows` out of `contributors` people."""
    preamble = PREAMBLE.format(
        count=f"{contributors:,}", commits=f"{total['commits']:,}", date=date.today().isoformat(), branch=BRANCH
    )
    if len(rows) < contributors:
        preamble += f"Only the top {len(rows):,} contributors by commit count are listed here.\n\n"
    table = ["| " + " | ".join(HEADERS) + " |", "|:---" + "|---:" * (len(HEADERS) - 1) + "|"]
    for name, commits, loc, files in rows:
        # escape the markdown table delimiter, names are commit-controlled input
        escaped = name.replace("|", "\\|")
        table.append(f"| {escaped} | {commits:,} | {loc:,} | {files:,} |")
    return preamble + "\n".join(table) + "\n"


def shorten(name):
    """Truncate `name` to what fits into the name column of the chart."""
    return name if len(name) <= SVG_NAME_MAX else name[: SVG_NAME_MAX - 1] + "…"


def render_svg(rows, contributors):
    """Return an SVG bar chart of `rows`, the busiest out of `contributors` people."""
    height = SVG_HEAD_H + len(rows) * SVG_ROW_H + SVG_FOOT_H
    bars = SVG_WIDTH - SVG_LABEL_W - SVG_COUNT_W - SVG_PAD
    most_commits = rows[0][1]
    out = [
        f'<svg xmlns="http://www.w3.org/2000/svg" width="{SVG_WIDTH}" height="{height}"'
        f' viewBox="0 0 {SVG_WIDTH} {height}" role="img"'
        f' aria-label="Borg contributors by commit count">',
        SVG_STYLE,
        f'<rect class="bg" x="0" y="0" width="{SVG_WIDTH}" height="{height}" rx="6"/>',
        f'<text class="title" x="{SVG_PAD}" y="26">Borg contributors</text>',
        f'<text class="subtitle" x="{SVG_PAD}" y="44">'
        f"top {len(rows)} of {contributors:,} by commits on the {BRANCH} branch</text>",
    ]
    for i, (name, commits, _loc, _files) in enumerate(rows):
        y = SVG_HEAD_H + i * SVG_ROW_H
        baseline = y + int(SVG_ROW_H * 0.7)
        # the scale is linear, but a bar should stay visible however short it is
        width = max(2, round(bars * commits / most_commits))
        # names are commit-controlled input, so & and < would break the XML
        out.append(f'<text class="name" x="{SVG_PAD}" y="{baseline}">{escape(shorten(name))}</text>')
        out.append(
            f'<rect class="bar" x="{SVG_LABEL_W}" y="{y + 5}" width="{width}" height="{SVG_ROW_H - 10}" rx="2"/>'
        )
        out.append(f'<text class="count" x="{SVG_WIDTH - SVG_PAD}" y="{baseline}">{commits:,}</text>')
    out.append(
        f'<text class="footer" x="{SVG_PAD}" y="{height - 10}">generated by scripts/fame.py using git-fame</text>'
    )
    out.append("</svg>")
    return "\n".join(out) + "\n"


def same_statistics(old, new):
    """Return True if the two FAME.md contents differ only in the generation date."""
    return DATE_LINE.sub("Generated on ", old) == DATE_LINE.sub("Generated on ", new)


def update(path, new, unchanged=str.__eq__):
    """Write `new` to `path` unless `unchanged` says it is already there."""
    if path.exists() and unchanged(path.read_text(encoding="utf-8"), new):
        print(f"{path} is up to date")
        return
    path.write_text(new, encoding="utf-8")
    print(f"wrote {path}")


def main():
    parser = argparse.ArgumentParser(description=__doc__.splitlines()[0])
    parser.add_argument("--top", type=int, default=0, metavar="N", help="only list the top N contributors")
    parser.add_argument(
        "--svg-top", type=int, default=SVG_TOP, metavar="N", help="chart the top N contributors [default: %(default)s]"
    )
    args = parser.parse_args()

    repo = Path(subprocess.check_output(["git", "rev-parse", "--show-toplevel"], text=True).strip())
    rows, total = collect()
    contributors = len(rows)
    people = [row for row in rows if not BOT.search(row[0])]
    if args.top:
        rows = rows[: args.top]
    print(f"{len(rows)} of {contributors} contributors, {contributors - len(people)} of them bots")
    update(repo / "FAME.md", render(rows, total, contributors), unchanged=same_statistics)
    update(repo / "FAME.svg", render_svg(people[: args.svg_top], len(people)))


if __name__ == "__main__":
    main()



import os
import re
import stat
import struct

from .posix import posix_acl_use_stored_uid_gid
from . import posix_ug
from ..helpers import workarounds
from ..helpers import safe_decode, safe_encode
from .base import SyncFile as BaseSyncFile
from .base import safe_fadvise
from .xattr import _listxattr_inner, _getxattr_inner, _setxattr_inner, split_string0
try:
    from .syncfilerange import sync_file_range, SYNC_FILE_RANGE_WRITE, SYNC_FILE_RANGE_WAIT_BEFORE, SYNC_FILE_RANGE_WAIT_AFTER
    SYNC_FILE_RANGE_LOADED = True
except ImportError:
    SYNC_FILE_RANGE_LOADED = False

from libc cimport errno



cdef extern from "sys/xattr.h":
    ssize_t c_listxattr "listxattr" (const char *path, char *list, size_t size)
    ssize_t c_llistxattr "llistxattr" (const char *path, char *list, size_t size)
    ssize_t c_flistxattr "flistxattr" (int filedes, char *list, size_t size)

    ssize_t c_getxattr "getxattr" (const char *path, const char *name, void *value, size_t size)
    ssize_t c_lgetxattr "lgetxattr" (const char *path, const char *name, void *value, size_t size)
    ssize_t c_fgetxattr "fgetxattr" (int filedes, const char *name, void *value, size_t size)

    int c_setxattr "setxattr" (const char *path, const char *name, const void *value, size_t size, int flags)
    int c_lsetxattr "lsetxattr" (const char *path, const char *name, const void *value, size_t size, int flags)
    int c_fsetxattr "fsetxattr" (int filedes, const char *name, const void *value, size_t size, int flags)

cdef extern from "sys/types.h":
    int ACL_TYPE_ACCESS
    int ACL_TYPE_DEFAULT

cdef extern from "sys/acl.h":
    ctypedef struct _acl_t:
        pass
    ctypedef _acl_t *acl_t

    int acl_free(void *obj)
    acl_t acl_get_file(const char *path, int type)
    acl_t acl_get_fd(int fd)
    int acl_set_file(const char *path, int type, acl_t acl)
    int acl_set_fd(int fd, acl_t acl)
    acl_t acl_from_text(const char *buf)

cdef extern from "acl/libacl.h":
    int acl_extended_file_nofollow(const char *path)
    int acl_extended_fd(int fd)
    char *acl_to_any_text(acl_t acl, const char *prefix, char separator, int options)
    int TEXT_NUMERIC_IDS

cdef extern from "linux/fs.h":
    # ioctls
    int FS_IOC_SETFLAGS
    int FS_IOC_GETFLAGS

    # inode flags
    int FS_NODUMP_FL
    int FS_IMMUTABLE_FL
    int FS_APPEND_FL

cdef extern from "sys/ioctl.h":
    int ioctl(int fildes, int request, ...)

cdef extern from "unistd.h":
    int _SC_PAGESIZE
    long sysconf(int name)

cdef extern from "string.h":
    char *strerror(int errnum)

_comment_re = re.compile(' *#.*', re.M)


def listxattr(path, *, follow_symlinks=False):
    def func(path, buf, size):
        if isinstance(path, int):
            return c_flistxattr(path, <char *> buf, size)
        else:
            if follow_symlinks:
                return c_listxattr(path, <char *> buf, size)
            else:
                return c_llistxattr(path, <char *> buf, size)

    n, buf = _listxattr_inner(func, path)
    return [name for name in split_string0(buf[:n])
            if name and not name.startswith(b'system.posix_acl_')]


def getxattr(path, name, *, follow_symlinks=False):
    def func(path, name, buf, size):
        if isinstance(path, int):
            return c_fgetxattr(path, name, <char *> buf, size)
        else:
            if follow_symlinks:
                return c_getxattr(path, name, <char *> buf, size)
            else:
                return c_lgetxattr(path, name, <char *> buf, size)

    n, buf = _getxattr_inner(func, path, name)
    return bytes(buf[:n])


def setxattr(path, name, value, *, follow_symlinks=False):
    def func(path, name, value, size):
        flags = 0
        if isinstance(path, int):
            return c_fsetxattr(path, name, <char *> value, size, flags)
        else:
            if follow_symlinks:
                return c_setxattr(path, name, <char *> value, size, flags)
            else:
                return c_lsetxattr(path, name, <char *> value, size, flags)

    _setxattr_inner(func, path, name, value)


BSD_TO_LINUX_FLAGS = {
    stat.UF_NODUMP: FS_NODUMP_FL,
    stat.UF_IMMUTABLE: FS_IMMUTABLE_FL,
    stat.UF_APPEND: FS_APPEND_FL,
}
# must be a bitwise OR of all values in BSD_TO_LINUX_FLAGS.
LINUX_MASK = FS_NODUMP_FL | FS_IMMUTABLE_FL | FS_APPEND_FL


def set_flags(path, bsd_flags, fd=None):
    if fd is None:
        st = os.stat(path, follow_symlinks=False)
        if stat.S_ISBLK(st.st_mode) or stat.S_ISCHR(st.st_mode) or stat.S_ISLNK(st.st_mode):
            # see comment in get_flags()
            return
    cdef int current
    cdef int flags
    cdef int mask = LINUX_MASK  # 1 at positions we want to influence
    cdef int new_flags = 0
    for bsd_flag, linux_flag in BSD_TO_LINUX_FLAGS.items():
        if bsd_flags & bsd_flag:
            new_flags |= linux_flag

    open_fd = fd is None
    if open_fd:
        fd = os.open(path, os.O_RDONLY|os.O_NONBLOCK|os.O_NOFOLLOW)
    try:
        # Get current flags.
        if ioctl(fd, FS_IOC_GETFLAGS, &current) == -1:
            # If this fails, give up because it is either not supported by the fs
            # or maybe not permitted? If we can't determine the current flags,
            # we better not risk corrupting them by setflags, see the comment below.
            return  # give up silently

        while True:
            # Replace only the bits we actually want to influence, keep others.
            # We can't just set all flags to the archived value, because we might
            # reset flags that are not controllable from userspace, see #9039.
            flags = (current & ~mask) | (new_flags & mask)

            if ioctl(fd, FS_IOC_SETFLAGS, &flags) != -1:
                return
            error_number = errno.errno
            # Usually we would only catch EOPNOTSUPP here, but Linux Kernel 6.17
            # has a bug where it returns ENOTTY instead of EOPNOTSUPP.
            if error_number in (errno.EOPNOTSUPP, errno.ENOTTY):
                return
            if error_number == errno.EPERM and mask != FS_NODUMP_FL:
                # Changing FS_IMMUTABLE_FL / FS_APPEND_FL needs CAP_LINUX_IMMUTABLE.
                # Retry, influencing only FS_NODUMP_FL, so an unprivileged extract
                # can still restore the nodump flag.
                mask = FS_NODUMP_FL
                continue
            raise OSError(error_number, strerror(error_number).decode(), path)
    finally:
        if open_fd:
            os.close(fd)


def get_flags(path, st, fd=None):
    if stat.S_ISBLK(st.st_mode) or stat.S_ISCHR(st.st_mode) or stat.S_ISLNK(st.st_mode):
        # Avoid opening device files — trying to open non-present devices can be rather slow.
        # Avoid opening symlinks; O_NOFOLLOW would make the open() fail anyway.
        return 0
    cdef int linux_flags
    open_fd = fd is None
    if open_fd:
        try:
            fd = os.open(path, os.O_RDONLY|os.O_NONBLOCK|os.O_NOFOLLOW)
        except OSError:
            return 0
    try:
        if ioctl(fd, FS_IOC_GETFLAGS, &linux_flags) == -1:
            return 0
    finally:
        if open_fd:
            os.close(fd)
    bsd_flags = 0
    for bsd_flag, linux_flag in BSD_TO_LINUX_FLAGS.items():
        if linux_flags & linux_flag:
            bsd_flags |= bsd_flag
    return bsd_flags


def acl_use_local_uid_gid(acl):
    """Replace the user/group field with the local uid/gid if possible
    """
    assert isinstance(acl, bytes)
    entries = []
    for entry in safe_decode(acl).split('\n'):
        if entry:
            fields = entry.split(':')
            if fields[0] == 'user' and fields[1]:
                fields[1] = str(posix_ug._user2uid(fields[1], fields[3]))
            elif fields[0] == 'group' and fields[1]:
                fields[1] = str(posix_ug._group2gid(fields[1], fields[3]))
            entries.append(':'.join(fields[:3]))
    return safe_encode('\n'.join(entries))


def _acl_from_numeric_to_named_with_id(acl):
    """Convert numeric-id ACL entries to name entries and append numeric id as 4th field.

    Input format (Linux libacl): lines like 'user:1000:rwx' or 'group:100:r-x' or 'user::rwx'.
    Output format: for entries with a name/id field, become 'user:uname:rwx:uid' or 'group:gname:r-x:gid'.
    """
    assert isinstance(acl, bytes)
    entries = []
    for entry in _comment_re.sub('', safe_decode(acl)).split('\n'):
        if not entry:
            continue
        fields = entry.split(':')
        # Expected 3 fields: type, ugid_or_empty, perms
        if len(fields) >= 3:
            typ, ugid_str, perm = fields[0], fields[1], fields[2]
            if ugid_str and typ == 'user':
                try:
                    uid = int(ugid_str)
                except ValueError:
                    uid = None
                uname = posix_ug._uid2user(uid, ugid_str) if uid is not None else ugid_str
                entries.append(':'.join([typ, uname, perm, str(uid if uid is not None else ugid_str)]))
            elif ugid_str and typ == 'group':
                try:
                    gid = int(ugid_str)
                except ValueError:
                    gid = None
                gname = posix_ug._gid2group(gid, ugid_str) if gid is not None else ugid_str
                entries.append(':'.join([typ, gname, perm, str(gid if gid is not None else ugid_str)]))
            else:
                # owner, group_obj, mask, other (empty ugid_str field) stay as-is
                entries.append(':'.join([typ, '', perm]))
        else:
            entries.append(entry)
    return safe_encode('\n'.join(entries))


def _acl_from_numeric_to_numeric_with_id(acl):
    """Keep numeric ids in name field and append the same id as 4th field where applicable."""
    assert isinstance(acl, bytes)
    entries = []
    for entry in _comment_re.sub('', safe_decode(acl)).split('\n'):
        if not entry:
            continue
        fields = entry.split(':')
        if len(fields) >= 3:
            typ, ugid, perm = fields[0], fields[1], fields[2]
            if ugid and (typ == 'user' or typ == 'group'):
                entries.append(':'.join([typ, ugid, perm, ugid]))
            else:
                entries.append(':'.join([typ, '', perm]))
        else:
            entries.append(entry)
    return safe_encode('\n'.join(entries))


def acl_get(path, item, st, numeric_ids=False, fd=None):
    cdef acl_t default_acl = NULL
    cdef acl_t access_acl = NULL
    cdef char *default_text = NULL
    cdef char *access_text = NULL
    cdef int ret = 0

    if isinstance(path, str):
        path = os.fsencode(path)
    if fd is not None:
        ret = acl_extended_fd(fd)
    else:
        ret = acl_extended_file_nofollow(path)
    if ret < 0:
        raise OSError(errno.errno, os.strerror(errno.errno), os.fsdecode(path))
    if ret == 0:
        # there is no ACL defining permissions other than those defined by the traditional file permission bits.
        # note: this should also be the case for symlink fs objects, as they can not have ACLs.
        return
    if numeric_ids:
        converter = _acl_from_numeric_to_numeric_with_id
    else:
        converter = _acl_from_numeric_to_named_with_id
    try:
        if fd is not None:
            access_acl = acl_get_fd(fd)
        else:
            access_acl = acl_get_file(path, ACL_TYPE_ACCESS)
        if access_acl == NULL:
            raise OSError(errno.errno, os.strerror(errno.errno), os.fsdecode(path))
        access_text = acl_to_any_text(access_acl, NULL, '\n', TEXT_NUMERIC_IDS)
        if access_text == NULL:
            raise OSError(errno.errno, os.strerror(errno.errno), os.fsdecode(path))
        item['acl_access'] = converter(access_text)
    finally:
        acl_free(access_text)
        acl_free(access_acl)
    if stat.S_ISDIR(st.st_mode):
        # only directories can have a default ACL. there is no fd-based api to get it.
        try:
            default_acl = acl_get_file(path, ACL_TYPE_DEFAULT)
            if default_acl == NULL:
                raise OSError(errno.errno, os.strerror(errno.errno), os.fsdecode(path))
            default_text = acl_to_any_text(default_acl, NULL, '\n', TEXT_NUMERIC_IDS)
            if default_text == NULL:
                raise OSError(errno.errno, os.strerror(errno.errno), os.fsdecode(path))
            item['acl_default'] = converter(default_text)
        finally:
            acl_free(default_text)
            acl_free(default_acl)


def acl_set(path, item, numeric_ids=False, fd=None):
    cdef acl_t access_acl = NULL
    cdef acl_t default_acl = NULL

    if stat.S_ISLNK(item.get('mode', 0)):
        # Linux does not support setting ACLs on symlinks
        return

    if isinstance(path, str):
        path = os.fsencode(path)
    if numeric_ids:
        converter = posix_acl_use_stored_uid_gid
    else:
        converter = acl_use_local_uid_gid
    access_text = item.get('acl_access')
    if access_text is not None:
        try:
            access_acl = acl_from_text(<bytes>converter(access_text))
            if access_acl == NULL:
                raise OSError(errno.errno, os.strerror(errno.errno), os.fsdecode(path))
            if fd is not None:
                if acl_set_fd(fd, access_acl) == -1:
                    raise OSError(errno.errno, os.strerror(errno.errno), os.fsdecode(path))
            else:
                if acl_set_file(path, ACL_TYPE_ACCESS, access_acl) == -1:
                    raise OSError(errno.errno, os.strerror(errno.errno), os.fsdecode(path))
        finally:
            acl_free(access_acl)
    default_text = item.get('acl_default')
    if default_text is not None:
        try:
            default_acl = acl_from_text(<bytes>converter(default_text))
            if default_acl == NULL:
                raise OSError(errno.errno, os.strerror(errno.errno), os.fsdecode(path))
            # only directories can get a default ACL. there is no fd-based api to set it.
            if acl_set_file(path, ACL_TYPE_DEFAULT, default_acl) == -1:
                raise OSError(errno.errno, os.strerror(errno.errno), os.fsdecode(path))
        finally:
            acl_free(default_acl)


# Linux kernel POSIX ACL xattr representation,
# see linux/include/uapi/linux/posix_acl_xattr.h and linux/include/linux/posix_acl.h.
_ACL_XATTR_VERSION = 2
_ACL_UNDEFINED_ID = 0xFFFFFFFF
# maps the ACL entry type to the e_tag values used for (unnamed, named) entries of that type.
_ACL_TAGS = {
    'user': (0x01, 0x02),   # ACL_USER_OBJ, ACL_USER
    'group': (0x04, 0x08),  # ACL_GROUP_OBJ, ACL_GROUP
    'mask': (0x10, None),   # ACL_MASK
    'other': (0x20, None),  # ACL_OTHER
}
_ACL_PERMS = {'r': 4, 'w': 2, 'x': 1, '-': 0}


def acl_text_to_xattr(acl, numeric_ids=False):
    """Convert an ACL from the borg item text representation (acl_access / acl_default)
    to the binary representation the Linux kernel uses for the system.posix_acl_access /
    system.posix_acl_default extended attributes.

    If `numeric_ids` is True, the stored numeric ids are used, otherwise the stored
    user/group names are mapped to local uids/gids (falling back to the stored ids).

    Raises ValueError for ACL text that can not be converted.
    """
    assert isinstance(acl, bytes)
    converter = posix_acl_use_stored_uid_gid if numeric_ids else acl_use_local_uid_gid
    entries = []
    try:
        for entry in safe_decode(converter(acl)).split('\n'):
            if not entry:
                continue
            typ, qualifier, perm = entry.split(':')[:3]
            unnamed_tag, named_tag = _ACL_TAGS[typ]
            if qualifier:
                if named_tag is None:
                    raise ValueError(f"unexpected qualifier in ACL entry: {entry}")
                tag, qid = named_tag, int(qualifier)
            else:
                tag, qid = unnamed_tag, _ACL_UNDEFINED_ID
            perms = 0
            for p in perm:
                perms |= _ACL_PERMS[p]
            entries.append((tag, perms, qid))
    except (IndexError, KeyError, ValueError) as e:
        raise ValueError(f"unsupported ACL: {acl!r}") from e
    # the kernel expects the entries in canonical order: owner, named users, owning group,
    # named groups, mask, other - with named entries sorted by id. the stored text should
    # already be in this order (libacl produces it), but do not rely on that.
    entries.sort(key=lambda e: (e[0], e[2]))
    return b''.join([struct.pack('<I', _ACL_XATTR_VERSION)] +
                    [struct.pack('<HHI', tag, perms, qid) for tag, perms, qid in entries])


cdef _sync_file_range(fd, offset, length, flags):
    assert offset & PAGE_MASK == 0, "offset %d not page-aligned" % offset
    assert length & PAGE_MASK == 0, "length %d not page-aligned" % length
    if sync_file_range(fd, offset, length, flags) != 0:
        raise OSError(errno.errno, os.strerror(errno.errno))
    safe_fadvise(fd, offset, length, 'DONTNEED')


cdef unsigned PAGE_MASK = sysconf(_SC_PAGESIZE) - 1


if 'basesyncfile' in workarounds or not SYNC_FILE_RANGE_LOADED:
    class SyncFile(BaseSyncFile):
        # if we are on platforms with a broken or not implemented sync_file_range,
        # use the more generic BaseSyncFile to avoid issues.
        # see basesyncfile description in our docs for details.
        pass
else:
    # a real Linux, so we can do better. :)
    class SyncFile(BaseSyncFile):
        """
        Implemented using sync_file_range for asynchronous write-out and fdatasync for actual durability.

        "write-out" means that dirty pages (= data that was written) are submitted to an I/O queue and will be send to
        disk in the immediate future.
        """

        def __init__(self, path, *, fd=None, binary=False):
            super().__init__(path, fd=fd, binary=binary)
            self.offset = 0
            self.write_window = (16 * 1024 ** 2) & ~PAGE_MASK
            self.last_sync = 0
            self.pending_sync = None

        def write(self, data):
            self.offset += self.f.write(data)
            offset = self.offset & ~PAGE_MASK
            if offset >= self.last_sync + self.write_window:
                self.f.flush()
                _sync_file_range(self.fd, self.last_sync, offset - self.last_sync, SYNC_FILE_RANGE_WRITE)
                if self.pending_sync is not None:
                    _sync_file_range(self.fd, self.pending_sync, self.last_sync - self.pending_sync,
                                     SYNC_FILE_RANGE_WRITE | SYNC_FILE_RANGE_WAIT_BEFORE | SYNC_FILE_RANGE_WAIT_AFTER)
                self.pending_sync = self.last_sync
                self.last_sync = offset

        def sync(self):
            self.f.flush()
            os.fdatasync(self.fd)
            # tell the OS that it does not need to cache what we just wrote,
            # avoids spoiling the cache for the OS and other processes.
            safe_fadvise(self.fd, 0, 0, 'DONTNEED')



# cython: language_level=3

from libc.stdint cimport uint8_t, uint64_t

from .base cimport ChunkerBase

# Shared chunker driver for the "UHF-then-PRF" (Chk-PHTE) chunkers: rabin-aes,
# goldilocks-aes and toeplitz-aes. They differ only in their rolling universal
# hash, which lives in their C kernels (see phte_core.h / phte_scan.h) and in
# the key derivation in their own modules. Buffering, min/max clamping,
# normalized chunking and the scan loop live in ChunkerBase; this class adds
# the 64-byte-window digest warm-up and the kernel-name plumbing.
#
# Subclasses implement _scan() and _digest64() by calling their C kernel. Both
# are Cython cdef methods, i.e. one vtable-dispatched call per *scan span*
# (up to a whole chunk of data) and per chunk start - never per byte, so the
# indirection is not measurable.


cdef class ChunkerPHTE(ChunkerBase):
    """
    Content-Defined Chunker, variable chunk sizes, UHF-then-PRF cut decision.

    Base class: a rolling universal hash compresses the last 64 bytes into a
    digest, AES-128 with a secret key is applied to it, and the cut decision
    only looks at the AES output. See the subclasses for the concrete
    universal hash and the security discussion.
    """

    cdef int _setup(self, str algo, int chunk_min_exp, int chunk_max_exp, int hash_mask_bits,
                    int nc_level, size_t normal_size, bint sparse) except 0:
        """Set up chunking parameters and the buffer (called by the subclass)."""
        # the rolling window needs 64 bytes of in-chunk history at the first
        # possible cut position (min_size), so the minimum chunk size must be
        # at least the window size:
        assert chunk_min_exp >= 6, "chunk_min_exp must be >= 6 (2^6 = 64 = window size)"
        self.kernel_str = "unknown"
        # the AES output is uniform, so contiguous low-bit masks are fine
        return self._setup_common(algo, chunk_min_exp, chunk_max_exp, hash_mask_bits,
                                  nc_level, normal_size, False, sparse)

    cdef uint64_t _digest64(self, const uint8_t *q) noexcept:
        """Digest of the 64 bytes at q (window warm-up). Implemented by subclasses."""
        return 0

    cdef uint64_t _restart(self) noexcept:
        # warm up the rolling digest over the 64 bytes preceding the first
        # scan position (they are within the current chunk, since min_size >= 64).
        return self._digest64(self.data + self.position - 64)

    @property
    def kernel(self):
        """Which scan kernel this chunker uses: 'aes-arm64', 'aes-ni' or 'evp'."""
        return self.kernel_str
